"""
Imperative style commands for constructing feedback in a convenient way.
Uses a global report object (MAIN_REPORT).
"""

__all__ = ['feedback', 'set_success', 'set_correct', 'compliment', 'give_partial', 'explain',
           'gently', 'hide_correctness', 'suppress', 'log', 'debug',
           'system_error', 'clear_report', 'get_all_feedback', 'guidance',
           'contextualize_report', 'Feedback', 'get_submission', 'set_formatter']

from pedal.core.feedback import (Feedback, FeedbackKind, FeedbackCategory,
                                 FeedbackResponse)
from pedal.core.report import MAIN_REPORT

from pedal.core.submission import Submission

#: Lowercase "function" version that works like other Core Feedback Functions.
feedback = Feedback


# TODO: force_success function, which ignores errors to give the points.

class set_correct(FeedbackResponse):
    """
    **(Feedback Function)**

    Creates Successful feedback for the user, indicating that the entire
    assignment is done.
    """
    title = "Complete"
    message_template = "Great work!"
    score = 1
    correct = True
    category = FeedbackCategory.COMPLETE
    kind = FeedbackKind.RESULT
    valence = Feedback.POSITIVE_VALENCE


set_success = set_correct


class compliment(FeedbackResponse):
    """
    Create a positive feedback for the user, potentially on a specific line of
    code.
    """
    category = FeedbackCategory.INSTRUCTOR
    kind = FeedbackKind.COMPLIMENT
    valence = Feedback.POSITIVE_VALENCE
    correct = True

    def __init__(self, message=None, title=None, message_template=None, **kwargs):
        if title is None:
            title = message
        if message is None and message_template is None:
            raise ValueError("compliment requires at least either message or message_template")
        super().__init__(message=message, title=title, message_template=message_template, **kwargs)


class give_partial(FeedbackResponse):
    """ Increases the user's current score by the `score`. """
    title = "Partial Credit"
    message_template = "Partial credit"
    category = FeedbackCategory.INSTRUCTOR
    kind = FeedbackKind.RESULT
    valence = Feedback.POSITIVE_VALENCE
    correct = None
    muted = True

    def __init__(self, value, **kwargs):
        super().__init__(score=value, **kwargs)


class explain(FeedbackResponse):
    """ Give a high-priority piece of negative feedback to the student. """
    title = "Instructor Feedback"
    muted = False
    category = Feedback.CATEGORIES.INSTRUCTOR
    kind = FeedbackKind.MISTAKE
    valence = Feedback.NEGATIVE_VALENCE

    def __init__(self, message=None, message_template=None, **kwargs):
        if message is None and message_template is None:
            raise ValueError("explain requires at least either message or message_template")
        super().__init__(message=message, message_template=message_template, **kwargs)


class gently(FeedbackResponse):
    """
    Give a low-priority piece of negative feedback to the student.

    Args:
        message (str): The feedback message to show to the student.

    """
    title = "Instructor Feedback"
    priority = Feedback.CATEGORIES.STUDENT
    muted = False
    category = Feedback.CATEGORIES.INSTRUCTOR
    kind = FeedbackKind.MISTAKE
    valence = Feedback.NEGATIVE_VALENCE

    def __init__(self, message=None, message_template=None, **kwargs):
        if message is None and message_template is None:
            raise ValueError("gently requires at least either message or message_template")
        super().__init__(message=message, message_template=message_template, **kwargs)


class guidance(FeedbackResponse):
    """ Give instructions about a question. """
    title = "Instructor Guidance"
    category = Feedback.CATEGORIES.INSTRUCTIONS
    kind = FeedbackKind.INSTRUCTIONAL
    valence = Feedback.NEUTRAL_VALENCE

    def __init__(self, message=None, message_template=None, **kwargs):
        if message is None and message_template is None:
            raise ValueError("compliment requires at least either message or message_template")
        super().__init__(message=message, message_template=message_template, **kwargs)


def hide_correctness(report=MAIN_REPORT):
    """
    Force the report to not indicate score/correctness.

    Args:
        report (:py:class:`pedal.core.report.Report`): The report object to
            hide correctness on.
    """
    report.hide_correctness()


def suppress(category=None, label=True, fields=None, report=MAIN_REPORT):
    """
    Hides a given category or label within a category from being considered
    by the resolver.

    Args:
        category (str): The general feedback category to suppress within.
            Should be a member of
            :py:class:`pedal.core.feedback_category.FeedbackCategory`.
        label (str or bool): The specific feedback label to suppress, or
            True if all the labels within this category should be suppressed.
        fields (dict): The fields that will be exactly matched to suppress a
            given feedback. The keys should be strings.
        report (:py:class:`~pedal.core.report.Report`): The report object to
            suppress information within.
    """
    report.suppress(category, label, fields)


def log(*items, sep=" ", **kwargs):
    """
    Attach logging information to the Report as a piece of feedback.

    Args:
        sep: The separator to use between items (defaults to space).
        items (Any): Any set of values to log information about. Will be
            converted to strings using `str` if not already strings.

    Returns:

    """
    kwargs.setdefault('category', Feedback.CATEGORIES.SYSTEM)
    kwargs.setdefault('muted', True)
    kwargs.setdefault('valence', Feedback.NEUTRAL_VALENCE)
    message = sep.join(item if isinstance(item, str) else str(item)
                       for item in items)
    feedback(message=message, label="log", **kwargs)


def debug(*items, **kwargs):
    """
    Attach logging information to the Report as a piece of feedback.
    Works at a higher priority than :py:func:`~pedal.core.commands.log` and
    does not attempt to convert to strings.

    TODO: Consider updating to match `log`

    Args:
        items (Any): Any set of values to log information about. Will be
            converted to strings using `str` if not already strings.

    Returns:

    """
    kwargs.setdefault('category', Feedback.CATEGORIES.SYSTEM)
    kwargs.setdefault('muted', True)
    kwargs.setdefault('priority', 'high')
    kwargs.setdefault('valence', Feedback.NEGATIVE_VALENCE)
    for item in items:
        item = item
        cloned_kwargs = kwargs.copy()
        cloned_kwargs.setdefault('message', item)
        feedback(label="debug", **kwargs)


def clear_report(report=MAIN_REPORT):
    """
    Removes all existing data from the report, including any submissions,
    suppressions, feedback, and Tool data.

    Args:
        report: The report to clear (defaults to the
            :py:data:`pedal.core.report.MAIN_REPORT`).
    """
    report.clear()


def contextualize_report(submission, filename='answer.py', clear=True,
                         report=MAIN_REPORT):
    """
    Updates the report with the submission. By default, clears out any old
    information in the report. You can pass in either an actual
    :py:class:`~pedal.core.submission.Submission` or a string representing
    the code of the submission.

    Args:
        submission (str or Submission):
        filename (str or None): If the `submission` was not a
            :py:class:`~pedal.core.submission.Submission`, then this will be
            used as the filename for the code given in ``submission``.
        clear (bool): Whether or not to clear the report before attaching
            the submission.
        report: The report to attach this feedback to (defaults to the
            :py:data:`~pedal.core.report.MAIN_REPORT`).
    """
    if not isinstance(submission, Submission):
        submission = Submission(files={filename: submission})
    if clear:
        report.clear()
    report.contextualize(submission)


def get_submission(report=MAIN_REPORT) -> Submission:
    """
    Get the current submission from the given report, or the default MAIN_REPORT.

    Args:
        report: The report to attach this feedback to (defaults to the
            :py:data:`~pedal.core.report.MAIN_REPORT`).

    Returns:
        Submission: The current submission
    """
    return report.submission


def get_all_feedback(report=MAIN_REPORT):
    """
    Gives access to the list of feedback from the report. Usually, you won't
    need this; but if you want to build on the results of earlier tools, it
    can be a useful mechanism.

    TODO: Provide mechanisms for conveniently searching feedback

    Args:
        report (:py:class:`~pedal.core.report.Report`): The report to attach
            this feedback to (defaults to the
            :py:data:`~pedal.core.report.MAIN_REPORT`).

    Returns:
        List[:py:class:`~pedal.core.feedback.Feedback`]: A list of feedback
            objects from the report.
    """
    return report.feedback


class system_error(FeedbackResponse):
    """
    Call this function to indicate that something has gone wrong at the system
    level with Pedal. Ideally, this doesn't happen, but sometimes errors
    cascade and its polite for tools to suggest that they are not working
    correctly. These will not usually be reported to the student.
    """
    title = "System Error"
    category = Feedback.CATEGORIES.SYSTEM
    kind = FeedbackKind.META
    valence = Feedback.NEUTRAL_VALENCE
    muted = True


# TODO: set_line_offset(offset, filename='answer.py')

def set_formatter(formatter, report=MAIN_REPORT):
    """
    Set the formatter for the given report.

    Args:
        formatter (Formatter): The formatter class to use. If you wish to use an instance instead,
            you'll need to call `set_formatter` on the report instance instead.
        report (:py:class:`~pedal.core.report.Report`): The report to attach
            this feedback to (defaults to the
            :py:data:`~pedal.core.report.MAIN_REPORT`).
    """
    report.set_formatter(formatter(report))


def set_pools(pools, report=MAIN_REPORT):
    """
    Ability to have A/B testing on a per-feedback basis.

    Args:
        pools (int or list[str]): Either the number of pools or the names of the pools. If a number is given,
            the pools are given the identifiers A/B/C/etc.
    """
    report.set_pools(pools)


"""
Tests for checking that the plotting extension works.
"""
import unittest
from textwrap import dedent

from pedal import call
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE
from pedal.extensions.plotting import *


code_hist_and_plot = ('''
import matplotlib.pyplot as plt
plt.hist([1,2,3])
plt.title("My line plot")
plt.show()
plt.plot([4,5,6])
plt.show()''')[1:]

code_xy_plot_wrong = ('''
import matplotlib.pyplot as plt
plt.plot([1,2,3], [4,5,6])
plt.title("My line plot")
plt.show()
''')


class TestPlots(ExecutionTestCase):

    def test_check_for_plot_correct_hist(self):
        with Execution(code_hist_and_plot) as e:
            self.assertEqual(assert_plot('hist', [1, 2, 3]), False)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_check_for_plot_wrong_hist(self):
        with Execution(code_hist_and_plot) as e:
            assert_plot('hist', [1, 2, 3, 4])
        self.assertFeedback(e, "Plot Data Incorrect\n"
                                "You have created a histogram, but it does not have the right data.\n"
                                "I expected the data to be:\n"
                                "    [1, 2, 3, 4]\n"
                                "But instead, the data was:\n"
                                "    [1, 2, 3]")

    def test_check_for_plot_correct_plot(self):
        with Execution(code_hist_and_plot) as e:
            self.assertEqual(assert_plot('line', [4, 5, 6]), False)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_check_for_plot_wrong_plot(self):
        with Execution(code_hist_and_plot) as e:
            assert_plot('line', [4, 5, 6, 7])
        self.assertFeedback(e, "Plot Data Incorrect\n"
                               "You have created a line plot, but it does not have the right data.\n"
                               "I expected the data to be:\n"
                               "    [4, 5, 6, 7]\n"
                               "But instead, the data was:\n"
                               "    [4, 5, 6]")

    @unittest.skip("Currently broken!")
    def test_check_for_xy_plot(self):
        with Execution(code_xy_plot_wrong) as e:
            assert_plot('line', [4, 5, 6])
        self.assertFeedback(e, "Plot Data Incorrect\n"
                               "You have created a line plot, but it does not have the right data.\n"
                               "I expected the data to be:\n"
                               "    [4, 5, 6]\n"
                               "But instead, the data was:\n"
                               "    [[1, 2, 3], [4, 5, 6]]")

    def test_assert_plot_wrong_type_of_plot(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.plot([1,2,3])
            plt.title("My line plot")
            plt.show()
        ''')
        with Execution(student_code) as e:
            assert_plot('hist', [1, 2, 3])
        self.assertFeedback(e, "Wrong Plot Type\n"
                               "You have plotted the right data, but you appear "
                                "to have not plotted it as a histogram.")

    def test_assert_plot_wrong_data_place(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.plot([1,2,3])
            plt.title("Wrong graph with the right data")
            plt.show()
            plt.hist([4,5,6])
            plt.title("Right graph with the wrong data")
            plt.show()
        ''')
        with Execution(student_code) as e:
            assert_plot('hist', [1, 2, 3])
        self.assertFeedback(e, "Plotting Another Graph\n"
                               "You have created a histogram, but it does not "
                               "have the right data. That data appears to have "
                               "been plotted in another graph.")

    def test_assert_plot_missing_plot_and_data(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.plot([1,2,3])
            plt.title("My line plot")
            plt.show()
        ''')
        with Execution(student_code) as e:
            assert_plot('hist', [4, 5, 6])
        self.assertFeedback(e, "Missing Plot\n"
                               "You have not created a histogram with the proper data.")

    def test_assert_plot_empty_scatter(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.scatter([], [])
            plt.title("Nothingness and despair")
            plt.show()
        ''')
        with Execution(student_code) as e:
            assert_plot('scatter', [])
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_plot_simple_scatter(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.scatter([1,2,3], [4,5,6])
            plt.title("Some actual stuff")
            plt.show()
        ''')
        with Execution(student_code) as e:
            assert_plot('scatter', [[1, 2, 3], [4, 5, 6]])
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_prevent_incorrect_plt(self):
        student_code = dedent('''
            import matplotlib.pyplot
            plt.scatter([1,2,3], [4,5,6])
            plt.title("Some actual stuff")
            plt.show()
        ''')
        with Execution(student_code) as e:
            self.assertTrue(prevent_incorrect_plt())
        self.assertFeedback(e, "Wrong MatPlotLib Import\n"
                               "You have imported the plt module, but"
                               " you did not rename it to "
                               "matplotlib.pyplot"
                               " using import matplotlib.pyplot as plt.")

    def test_prevent_incorrect_plt_missing_plt(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            scatter([1,2,3], [4,5,6])
            plt.title("Some actual stuff")
            plt.show()
        ''')
        with Execution(student_code) as e:
            self.assertTrue(prevent_incorrect_plt())
        self.assertFeedback(e, "Missing MatPlotLib Import\n"
                               "You have attempted to use the MatPlotLib "
                               "function named plt.scatter. "
                               "However, you imported MatPlotLib in a way that "
                               "does not allow you to use the function "
                               "directly. I recommend you use "
                               "scatter instead, after you use "
                               "import matplotlib.pyplot as plt.")

    def test_prevent_incorrect_plt_success(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.scatter([1,2,3], [4,5,6])
            plt.title("Some actual stuff")
            plt.show()
        ''')
        with Execution(student_code) as e:
            self.assertFalse(prevent_incorrect_plt())


    def test_check_plot_wrong_scatter(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.scatter([1,2,3, 4], [4,5,6, 7])
            plt.show()
        ''')
        with Execution(student_code) as e:
            assert_plot('scatter', [[1, 2, 3], [4, 5, 6]])
        self.assertFeedback(e, "Plot Data Incorrect\n"
                               "You have created a scatter plot, but it does not have the right data.\n"
                               "I expected the data to be:\n"
                               "    [[1, 2, 3], [4, 5, 6]]\n"
                               "But instead, the data was:\n"
                               "    [[1, 2, 3, 4], [4, 5, 6, 7]]")

    def test_check_plot_wrong_scatter_via_call(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            def make_plot():
                plt.scatter([1,2,3, 4], [4,5,6, 7])
                plt.show()
            make_plot
            plt
        ''')
        with Execution(student_code) as e:
            assert_plot('scatter', [[1, 2, 3], [4, 5, 6]], context=call('make_plot'))
        self.assertFeedback(e, "Plot Data Incorrect\n"
                               "You have created a scatter plot, but it does not have the right data.\n"
                               " I ran the code:\n    make_plot()\n\n"
                               "I expected the data to be:\n"
                               "    [[1, 2, 3], [4, 5, 6]]\n"
                               "But instead, the data was:\n"
                               "    [[1, 2, 3, 4], [4, 5, 6, 7]]")

    def test_check_plot_custom_compare(self):
        student_code = dedent('''
            import matplotlib.pyplot as plt
            plt.hist([82.0, 70.0, 57.99999999999999, 59.0, 83.0, 81.0, 56.99999999999999])
            plt.show()
        ''')
        def close_comparison(a, b):
            return set(round(aa) for aa in a) == set(round(bb) for bb in b)
        with Execution(student_code) as e:
            assert_plot('hist', [82.0, 70.0, 58.0, 59.0, 83.0, 81.0, 57.0],
                        special_comparison=close_comparison)
        self.assertFeedback(e, "Complete\nGreat work!")


"""
Collection of commonly reusable positive feedback. Let's all be a little better!
"""
from pedal.core.commands import compliment


class correct_output(compliment):
    """ Signals that the student's output looks correct. """
    message = "Your output looks correct."
    title = "Correct Output"

    def __init__(self, **kwargs):
        super().__init__(self.message, **kwargs)


class close_output(compliment):
    message = "Your output is close to being correct!"
    title = "Output Is Close"

    def __init__(self, reason, **kwargs):
        fields = kwargs.setdefault('fields', {})
        fields['reason'] = reason
        super().__init__(self.message, **kwargs)



import unittest
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pedal.core.commands import *


class TestCode(unittest.TestCase):

    def test_gently(self):
        clear_report()
        gently('You should always create unit tests.')

        self.assertEqual(len(get_all_feedback()), 1)


    def test_guidance(self):
        clear_report()
        guidance("Make sure you do the next steps")

        self.assertEqual(len(get_all_feedback()), 1)


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Class for managing custom score objects (e.g., "+20%").
"""
import re

SCORE_PATTERN = re.compile(r"(!*)([+\-/*])?([\d.]+)(%)?(.*)")

class Score:
    def __init__(self, invert, operator, value, percentage, leftovers):
        self.invert = invert
        self.operator = operator
        self.value = value
        self.percentage = percentage
        self.leftovers = leftovers

    @classmethod
    def parse(cls, score):
        match = SCORE_PATTERN.match(score)
        if not match:
            # TODO: Add context of feedback being processed
            raise ValueError(f"Invalid Score string: {score}")
        invert = bool(len(match.group(1)) % 2)
        operator = match.group(2)
        value = float(match.group(3))
        percentage = bool(match.group(4) == "%")
        leftovers = match.group(5)
        if percentage:
            value = value / 100.0
        return Score(invert, operator, value, percentage, leftovers)

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return Score(self.invert, self.operator, self.value/other, self.percentage, self.leftovers)
        return NotImplemented(f"Division is not supported between a Score and a {type(other)}")

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Score(self.invert, self.operator, self.value*other, self.percentage, self.leftovers)
        return NotImplemented(f"Multiplication is not supported between a Score and a {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __add__(self, other):
        if isinstance(other, (int, float)):
            return Score(self.invert, self.operator, self.value+other, self.percentage, self.leftovers)
        return NotImplemented(f"Addition is not supported between a Score and a {type(other)}")

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            return Score(self.invert, self.operator, self.value-other, self.percentage, self.leftovers)
        return NotImplemented(f"Subtraction is not supported between a Score and a {type(other)}")

    def __str__(self):
        value = self.value
        if self.percentage:
            # TODO: Being careless with rounding here.
            value *= 100
        value = str(round(value))
        operator = self.operator
        if not operator:
            operator = "+"
        return (
            ("!" if self.invert else "")+operator+value +
            ("%" if self.percentage else "")+self.leftovers
        )

    def to_percent_string(self):
        value = self.value * 100
        value = str(round(value))
        operator = "+" if not self.operator else self.operator
        return ("!" if self.invert else "") + operator + value + "%" + self.leftovers

    def add_to_current(self, current):
        if self.operator in ("+", None, "") and not self.invert:
            current += self.value
        elif self.operator == "-" and not self.invert:
            current -= self.value
        elif self.operator == "*" and not self.invert:
            current *= self.value
        elif self.operator == "/" and not self.invert:
            current /= self.value
        return current


def combine_scores(scores, invert_scores=None) -> float:
    """
    Combines the scores (and possible invert_scores) into a single number.
    Args:
        scores ():
        invert_scores ():

    Returns:
        total_score (float): Two decimal place float of final result.
    """
    total = 0
    for score in scores:
        if isinstance(score, (int, float)):
            total += score
        elif isinstance(score, str):
            total = Score.parse(score).add_to_current(total)
    if invert_scores:
        for score in invert_scores:
            if isinstance(score, str):
                total = Score.parse("!"+score).add_to_current(total)
    return round(total, 2)



"""
A package for analyzing student code.
"""
import sys
import os
import logging

# Setup Logging
logger = logging.getLogger('pedal')
logger.addHandler(logging.NullHandler())
logging.basicConfig(
    level=os.environ.get('LOGLEVEL', 'ERROR').upper()
)

logger.debug("Pedal import started")

# Core Features
from pedal.core.report import MAIN_REPORT
from pedal.core.submission import Submission
from pedal.core.commands import *
logger.debug("Pedal core import finished")

# Default Tools
from pedal.source import *
logger.debug("Source Tool import finished")

from pedal.sandbox.commands import *
logger.debug("Sandbox Tool import finished")

from pedal.cait import *
logger.debug("CAIT Tool import finished")

from pedal.assertions.commands import *
logger.debug("Assertions Tool import finished")

from pedal.resolvers import *
logger.debug("Resolvers Tool import finished")

# Provide type of `student` data object (setup via Sandbox)
student: Sandbox

logger.debug("Pedal import finished")



from pedal.types.builtin import BUILTIN_MODULES
from pedal.types.new_types import ModuleType, FunctionType, register_builtin_module
from pedal.types.new_types import void_function, int_function, ClassType, bool_function, InstanceType

_VOID_FUNCTIONS = ['run_every', 'panic', 'sleep', 'set_volume']


def load_microbit_module():
    _BUTTON_CLASS = ClassType('Button', {
        'is_pressed': bool_function('is_pressed'),
        'was_pressed': bool_function('was_pressed'),
        'get_presses': int_function('get_presses')
    }, [])

    _MICROBIT_FIELDS = {
        'running_time': int_function('running_time'),
        'temperature': int_function('temperature'),
        'Button': _BUTTON_CLASS,
        'button_a': InstanceType(_BUTTON_CLASS),
        'button_b': InstanceType(_BUTTON_CLASS),
    }
    _MICROBIT_FIELDS.update({
        name: void_function for name in _VOID_FUNCTIONS
    })

    _MICROBIT_SUBMODULES = {
        'display': ModuleType('display', {
            'show': void_function('show'),
            'scroll': void_function('scroll'),
            'on': void_function('on'),
            'off': void_function('off'),
            'clear': void_function('clear'),
            'set_pixel': void_function('set_pixel'),
            'get_pixel': int_function('get_pixel'),
            'is_on': bool_function('is_on'),
            'read_light_level': int_function('read_light_level')
        })
    }

    return ModuleType('microbit', fields=_MICROBIT_FIELDS, submodules=_MICROBIT_SUBMODULES)


register_builtin_module('microbit', load_microbit_module)



"""
Simple data classes for storing feedback to present to learners.
"""

__all__ = ['Feedback', 'FeedbackKind', 'FeedbackCategory',
           "CompositeFeedbackFunction",
           "FeedbackResponse"]

from pedal.core.formatting import wrap_fields
from pedal.core.location import Location
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback_category import FeedbackKind, FeedbackCategory, FeedbackStatus

PEDAL_DEVELOPERS = ["Austin Cory Bart <acbart@udel.edu>",
                    "Luke Gusukuma <lukesg08@vt.edu>"]


class FeedbackRegistry:
    def __init__(self):
        self._registered_feedback = {}


class Feedback:
    """
    A class for storing raw feedback.

    Attributes:
        label (str): An internal name for this specific piece of feedback. The
            label should be an underscore-separated string following the same
            conventions as names in Python. They do not have to be globally
            unique, but labels should be as unique as possible (especially
            within a category).
        tool (str, optional): An internal name for indicating the tool that created
            this feedback. Should be taken directly from the Tool itself. If `None`, then
            this was not created by a tool but directly by the control script.
        category (str): A human-presentable name showable to the learner, indicating what
            sort of feedback this falls into (e.g., "runtime", "syntax", "algorithm").
            More than one feedback will be in a category, but a feedback cannot be in
            more than one category.
        kind (str): The pedagogical role of this feedback, e.g., "misconception", "mistake",
            "hint", "constraint". Usually, a piece of Feedback is pointing out a mistake,
            but feedback can also be used for various other purposes.
        justification (str): An instructor-facing string briefly describing why this
            feedback was selected. Serves as a "TL;DR" for this feedback category, useful
            for debugging why a piece of feedback appeared.
        justification_template (str): A markdown-formatted message template that will
            be used if a ``justification`` is None. Any ``fields`` will be injected
            into the template IF the ``condition`` is met.
        priority (str): An indication of how important this feedback is relative to other types
            of feedback in the same category. Might be "high/medium/low". Exactly how this
            gets used is up to the resolver, but typically it helps break ties.
        valence (int): Indicates whether this is negative, positive, or neutral feedback.
            Either 1, -1, or 0.

        title (str, optional): A formal, student-facing title for this feedback. If None, indicates
            that the :py:attr:`~pedal.core.feedback.Feedback.label` should be used instead.
        message (str): A markdown-formatted message (aka also supporting HTML) that could be rendered
            to the user.
        message_template (str): A markdown-formatted message template that will
            be used if a ``message`` is None. Any ``fields`` will be injected
            into the template IF the ``condition`` is met.
        fields (Dict[str,Any]): The raw data that was used to interpolate the
            template to produce the message.
        location (:py:attr:`~pedal.core.location.Location` or int): Information
            about specific locations relevant to this message.

        score (int): A numeric score to modify the students' total score, indicating their overall performance.
            It is ultimately up to the Resolver to decide how to combine all the different scores; a typical
            strategy would be to add all the scores together for any non-muted feedback.
        correct (bool): Indicates that the entire submission should be considered correct (success) and that the
            task is now finished.
        muted (bool): Whether this piece of feedback is something that should be shown to a student. There are
            various use cases for muted feedback: they can serve as flags for later conditionals, suppressed
            default kinds of feedback, or perhaps feedback that is interesting for analysis but not pedagogically
            helpful to give to the student. They will still contribute to overall score, but not to the correctness
            of the submission.
        unscored (bool): Whether this piece of feedback contributes to the score/correctness.

        else_message (str): A string to render as a message when a
            NEGATIVE valence feedback is NOT triggered, or a POSITIVE valence
            feedback IS triggered.
        else_message_template (str): Similar to the ``message_template``, but for the ``else_message``.

        activate (bool): Used for default feedback objects without a custom
            condition, to indicate whether they should be considered triggered.
            Defaults to True; setting this to False means that the feedback
            object will be deactivated. Note that most inheriting Feedback
            Functions will not respect this parameter.

        author (List[str]): A list of names/emails that indicate who created this piece of feedback. They can be
            either names, emails, or combinations in the style of `"Cory Bart <acbart@udel.edu>"`.
        version (str): A version string in the style of Semantic Version (semvar) such as `"0.0.1"`. The last (third)
            digit should be incremented for small bug fixes/changes. The middle (second) digit should be used for more
            serious and intense changes. The first digit should be incremented when changes are made on exposure to
            learners or some other evidence-based motivation.
        tags (list[:py:class:`~pedal.core.tag.Tag`]): Any tags that you want to
            attach to this feedback.

        parent (int, str, or `pedal.core.feedback.Feedback`): Information about
            what logical grouping within the submission this belongs to.
            Various tools can chunk up a submission (e.g., by section), they
            can use this field to keep track of how that decision was made.
            Resolvers can also use this information to organize feedback or to
            report multiple categories.
        report (:py:class:`~pedal.core.report.Report`): The Report object to
            attach this feedback to. Defaults to MAIN_REPORT. Unspecified fields
            will be filled in by inspecting the current Feedback Function
            context.
    """

    POSITIVE_VALENCE = 1
    NEUTRAL_VALENCE = 0
    NEGATIVE_VALENCE = -1
    CATEGORIES = FeedbackCategory
    KINDS = FeedbackKind

    DEFAULT_FEEDBACK_MESSAGE = "No feedback message provided"
    DEFAULT_JUSTIFICATION_MESSAGE = "No justification provided"
    DEFAULT_ELSE_MESSAGE = None

    label = None
    category = None
    justification = None
    justification_template = None
    constant_fields = None
    fields = None
    field_names = None
    kind = None
    title = None
    message = None
    message_template = None
    else_message = None
    else_message_template = None
    priority = None
    valence = None
    location = None
    score = None
    correct = None
    muted = None
    unscored = None
    tool = None
    version = '1.0.0'
    author = PEDAL_DEVELOPERS
    tags = None
    parent = None

    activate: bool
    _status: str
    _exception: Exception or None
    _met_condition: bool
    _stored_args: tuple
    _stored_kwargs: dict
    _override_backups = None
    _pools = {}

    resolved_score = None
    unused_message = None

    # MAIN_REPORT

    def __init__(self, *args, label=None,
                 category=None, justification=None,
                 fields=None, field_names=None,
                 kind=None, title=None,
                 message=None, message_template=None,
                 else_message=None, else_message_template=None,
                 priority=None, valence=None,
                 location=None, score=None, correct=None,
                 muted=None, unscored=None,
                 tool=None, version=None, author=None,
                 tags=None, parent=None, report=MAIN_REPORT,
                 delay_condition=False, activate=True,
                 **kwargs):
        # Internal data
        self.report = report
        # Metadata
        if label is not None:
            self.label = label
        else:
            self.label = self.__class__.__name__
        # Condition
        if category is not None:
            self.category = category
        if justification is not None:
            self.justification = justification
        self.activate = activate
        # Response
        if kind is not None:
            self.kind = kind
        if priority is not None:
            self.priority = priority
        if valence is not None:
            self.valence = valence

        # Presentation
        if fields is not None:
            self.fields = fields
        else:
            self.fields = {}
        if self.constant_fields is not None:
            self.fields.update(self.constant_fields)
        if field_names is not None:
            self.field_names = field_names
            # TODO: Should this be taken from `fields` if nothing is provided?
        if title is not None:
            self.title = title
        elif self.title is None:
            self.title = label
        if message is not None:
            self.message = message
        if message_template is not None:
            self.message_template = message_template
        if else_message is not None:
            self.else_message = else_message
        if else_message_template is not None:
            self.else_message_template = else_message_template

        # Locations
        if isinstance(location, int):
            location = Location(location)
        # TODO: Handle tuples (Line, Col) and (Filename, Line, Col), and
        #  possibly lists thereof
        if location is not None:
            self.location = location
        # Result
        if score is not None:
            self.score = score
        if correct is not None:
            self.correct = correct
        if muted is not None:
            self.muted = muted
        if unscored is not None:
            self.unscored = unscored
        # Metadata
        if tool is not None:
            self.tool = tool
        if version is not None:
            self.version = version
        if author is not None:
            self.author = author
        if tags is not None:
            self.tags = tags
        # Organizational
        if parent is not None:
            self.parent = parent
        if self.parent is None:
            # Might inherit from Report's current group
            self.parent = self.report.get_current_group()
        if self.field_names is not None:
            for field_name in self.field_names:
                self.fields[field_name] = kwargs.get(field_name)
        for key, value in kwargs.items():
            self.fields[key] = value
        if 'location' not in self.fields and self.location is not None:
            self.fields['location'] = self.location
        # Potentially delay the condition check
        self._stored_args = args
        self._stored_kwargs = kwargs
        if delay_condition:
            self._met_condition = False
            self._status = FeedbackStatus.DELAYED
        else:
            self._handle_condition()

    def _handle_condition(self):
        """ Actually handle the condition check, updating message and report. """
        # Self-attach to a given report?
        self._exception = None
        try:
            self._met_condition = self.condition(*self._stored_args, **self._stored_kwargs)
            # Generate the message field as needed
            self.justification = self._get_justification(self._met_condition)
            if self._met_condition:
                self.message = self._get_message()
                self._status = FeedbackStatus.ACTIVE
            else:
                self.else_message = self._get_else_message()
                self.message = self.else_message
                try:
                    self.unused_message = self._get_message()
                except Exception:
                    self.unused_message = ""
                self._status = FeedbackStatus.INACTIVE
        except Exception as e:
            self._met_condition = False
            self._exception = e
            self._status = FeedbackStatus.ERROR
        if self.report is not None:
            if self._met_condition:
                self.report.add_feedback(self)
            else:
                self.report.add_ignored_feedback(self)
        # Free up these temporary fields after condition is handled
        # del self._stored_args
        # del self._stored_kwargs
        if self._exception is not None:
            raise self._exception

    def condition(self, *args, **kwargs):
        """
        Detect if this feedback is present in the code.
        Defaults to true through the `activate` parameter.

        Returns:
            bool: Whether this feedback's condition was detected.
        """
        return self.activate

    def _get_message(self):
        """
        Determines the appropriate value for the message. It will attempt
        to use this instance's message, but if it's not available then it will
        try to generate one from the message_template. Then, it returns a
        generic message.

        You can override this to create a truly dynamic message, if you want.

        Returns:
            str: The message for this feedback.
        """
        if self.message is not None:
            return self.message
        if self.message_template is not None:
            fields = wrap_fields(self.report.format, self.fields)
            return self.message_template.format(**fields)
        return self.DEFAULT_FEEDBACK_MESSAGE

    def _get_else_message(self):
        """
        Determines the appropriate value for the else_message. It will attempt
        to use this instance's else_message, but if it's not available then it will
        try to generate one from the else_message_template. Then, it returns ``None``
        instead (since usually we don't provide positive feedback).

        You can override this to create a truly dynamic else_message, if you want.

        Returns:
            str: The else_message for this feedback.
        """
        if self.else_message is not None:
            return self.else_message
        if self.else_message_template is not None:
            fields = wrap_fields(self.report.format, self.fields)
            return self.else_message_template.format(**fields)
        return self.DEFAULT_ELSE_MESSAGE

    def _get_justification(self, met_condition):
        """
        Determines the appropriate value for the justification. It first checks
        if there is a `justification` already present, but if it's not available
        then it will attempt to generate one from the `justification_template`. Then,
        it returns a generic message.

        You can override this to create a truly dynamic justification, if you want.

        Returns:
            str: The justification for this feedback.
        """
        if self.justification is not None:
            if isinstance(self.justification, str):
                if met_condition:
                    return self.justification
                else:
                    return "The following condition was not met: " + self.justification
            else:
                met, unmet = self.justification
                return met if met_condition else unmet
        if self.justification_template is not None:
            if isinstance(self.justification_template, str):
                template = "The following condition was not met: " + self.justification_template
            else:
                met, unmet = self.justification_template
                template = met if met_condition else unmet
            fields = wrap_fields(self.report.format, self.fields)
            return template.format(**fields)
        return self.DEFAULT_JUSTIFICATION_MESSAGE

    def _get_child_feedback(self, feedback, active):
        """ Callback function that Reports will call when a new piece of
        feedback is being considered. By default, does nothing. This is useful
        for :py:class:`pedal.core.group.FeedbackGroup`, most other feedbacks
        will just not care.

        The ``active`` parameter controls whether the condition for the
        feedback was met. """

    def __xor__(self, other):
        """
        Allows you to have two mutually exclusive conditions. Only one condition will be activated.

        Args:
            other:

        Returns:

        """
        if isinstance(other, Feedback):
            self.muted = bool(self) and not bool(other)
            self.unscored = self.muted
            other.muted = bool(other) and not bool(self)
            other.unscored = other.muted
        if isinstance(other, bool):
            self.muted = bool(self) and not bool(other)
            self.unscored = self.muted

    def __rxor__(self, other):
        return self.__xor__(other)

    def __bool__(self):
        return bool(self._met_condition)

    def __str__(self):
        """
        Create a string representation of this piece of Feedback for quick, dev purposes.
        Returns:
            str: String representation
        """
        return "<Feedback ({},{})>".format(self.category, self.label)

    def __repr__(self):
        """
        Create a string representation of this piece of Feedback that displays considerably more information.
        Returns:
            str: String representation with more data
        """
        metadata = ""
        if self.tool is not None:
            metadata += ", tool=" + self.tool
        if self.category is not None:
            metadata += ", category=" + self.category
        if self.priority is not None:
            metadata += ", priority=" + self.priority
        if self.parent is not None:
            metadata += ", parent=" + str(self.parent.label)
        if self.fields is not None:
            metadata += ", " + ", ".join(f"{field}={value!r}" for field, value in self.fields.items())
        return "Feedback({}{})".format(self.label, metadata)

    def update_location(self, location):
        """ Updates both the fields and location attribute.
        TODO: Handle less information intelligently. """
        if isinstance(location, int):
            location = Location(location)
        self.location = location
        self.fields['location'] = location

    def _fields_to_json(self):
        return self.fields.copy()

    def to_json(self):
        return {
            'correct': self.correct,
            'score': self.score,
            'title': self.title,
            'message': self.message,
            'label': self.label,
            'active': bool(self),
            'muted': self.muted,
            'unscored': self.unscored,
            'category': self.category,
            'kind': self.kind,
            'valence': self.valence,
            'version': self.version,
            'fields': self._fields_to_json(),
            'justification': self.justification,
            'priority': self.priority,
            'location': self.location.to_json() if self.location is not None else None
        }

    @classmethod
    def override(cls, report=MAIN_REPORT, **fields):
        if cls._override_backups is None:
            cls._override_backups = {}
        for field, new_value in fields.items():
            if field not in cls._override_backups:
                cls._override_backups[field] = getattr(cls, field)
            setattr(cls, field, new_value)
        report.override_feedback(cls)

    @classmethod
    def _restore_overrides(cls):
        for field, old_value in cls._override_backups.items():
            setattr(cls, field, old_value)
        cls._override_backups.clear()


    @classmethod
    def override_for_pool(cls, pool, **fields):
        if isinstance(pool, str):
            pool = [pool]
        for each_pool in pool:
            if each_pool not in cls._pools:
                cls._pools[each_pool] = {}
            cls._pools[each_pool].update(fields)

    def _finalize(self):
        # TODO: Move this to be earlier, when we create a feedback.
        #       Otherwise there might be fields that are not interpolated correctly!
        possible_overrides = self._pools.get(self.report.chosen_pool)
        if possible_overrides:
            for key, value in possible_overrides.items():
                setattr(self, key, value)


class FeedbackResponse(Feedback):
    """
    An extension of :py:class:`~pedal.core.feedback.Feedback` that is meant
    to indicate that the class should not have any condition behind its
    response.
    """


def CompositeFeedbackFunction(*functions):
    """
    Decorator for functions that return multiple types of feedback functions.

    Args:
        functions (callable): A list of callable functions.

    Returns:
        callable: The decorated function.
    """

    def CompositeFeedbackFunction_with_attrs(function):
        """

        Args:
            function:

        Returns:

        """
        CompositeFeedbackFunction_with_attrs.functions = functions
        return function

    return CompositeFeedbackFunction_with_attrs


class FeedbackGroup(Feedback):
    """
    An extension of :py:class:`~pedal.core.feedback.Feedback` that is meant
    to indicate that this class will start a new Group context within the report
    and do something interesting with any children it gets.
    """


DEFAULT_CATEGORY_PRIORITY = [
    "highest",
    # Static
    Feedback.CATEGORIES.SYNTAX,
    Feedback.CATEGORIES.MISTAKES,
    Feedback.CATEGORIES.INSTRUCTOR,
    Feedback.CATEGORIES.ALGORITHMIC,
    # Dynamic
    Feedback.CATEGORIES.RUNTIME,
    Feedback.CATEGORIES.STUDENT,
    Feedback.CATEGORIES.SPECIFICATION,
    Feedback.CATEGORIES.POSITIVE,
    Feedback.CATEGORIES.INSTRUCTIONS,
    Feedback.CATEGORIES.UNKNOWN,
    "lowest"
]



"""
Collected utility functions for across Pedal.

To decide if something belongs here, ask if multiple Tools might want to use it,
possibly including Report itself. Ask if it's something you'd probably find
in a separate module, rather than being a strictly pedagogical tool. All of
these are evidence that it belongs here.

Is it something that an Instructor might want to change? Then it probably
belongs in Report or something tool specific. But you might want to abstract
away parts into here and parts into Report.
"""


"""
This module contains AstSymbol, AstSymbolList, and AstMap, three classes
that are used to model names and matched expressions.
"""
from pedal.cait.cait_node import CaitNode
from pedal.core.location import Location

SymTables = {"FUNC": "add_func_to_sym_table",
             "VAR": "add_var_to_sym_table",
             "CLASS": "add_class_to_sym_table"}


class AstSymbol:
    """
    This represents an Ast symbol, whether it be a variable (name node) or a function name
    for place holders used in instructor patterns

    Notes:
        Also has the attributes of the relevant Name node from the ast class.

    Attributes:
        id (str): the name of the variable place holder used by the instructor
        ast_node (cait_node): the ast node of the variable
    """

    def __init__(self, _id="", _node=None):
        self.id = _id
        self.astNode = _node
        self.ast_node = _node

    def __getattr__(self, attr):
        return getattr(self.astNode, attr)

    def __str__(self):
        # return ''.join(["id = ", self.id.__str__(), ", astNode = ", type(self.astNode).__name__])
        return self.id

    def __repr__(self):
        return ''.join(["id = ", self.id.__str__(), ", astNode = ", type(self.astNode).__name__])


class AstSymbolList:
    """
    This class is a wrapper for a list of AstSymbols for ease of access
    If accessed as a list, manipulable as a list, otherwise, acts as the first AstSymbol in the list
    """

    def __init__(self):
        self.my_list = []

    def __getitem__(self, item):
        return self.my_list.__getitem__(item)

    def append(self, item):
        """

        Args:
            item:
        """
        self.my_list.append(item)

    def __getattr__(self, attr):
        return getattr(self.my_list[0], attr)

    def __len__(self):
        return self.my_list.__len__()


class AstMap:
    """
    TODO: Seriously think about redoing this mapping system.
    """
    def __init__(self):
        self.mappings = {}
        self.symbol_table = {}
        self.exp_table = {}
        self.func_table = {}
        self.class_table = {}
        self.conflict_keys = []
        self.match_root = None
        self.diagnosis = ""

    def add_x_to_sym_table(self, key, value, x_table):
        if key in x_table:
            new_list = x_table[key]
            if value not in new_list:
                new_list.append(value)
            if not (key in self.conflict_keys):
                for other in new_list:
                    if value.id != other.id:
                        self.conflict_keys.append(key)
                        break
        else:
            new_list = AstSymbolList()
            new_list.append(value)

        x_table[key] = new_list
        return len(self.conflict_keys)

    def add_class_to_sym_table(self, ins_node, std_node):
        """
        Adds ins_node.name to the symbol table if it doesn't already exist, mapping it to a set of ins_node. Updates a
        second dictionary that maps ins_node to an std_node, and overwrites the current std_node since there should only
        be one mapping.

        Args:
            ins_node: instructor node or str representing a class name
            std_node: student node representing function

        Returns:
            int: number of conflicts generated

        """
        if not isinstance(std_node, CaitNode):
            raise TypeError
        if isinstance(ins_node, str):
            key = ins_node
        else:
            try:
                if ins_node.ast_name == "ClassDef":
                    key = ins_node.astNode.name
                else:  # TODO: Little skulpt artifact that doesn't raise Attribute Errors...
                    key = ins_node._id
                    raise AttributeError
            except AttributeError:
                key = ins_node.astNode._id

        if std_node.ast_name == "ClassDef":
            value = AstSymbol(std_node.astNode.name, std_node)
        else:  # TODO: Little skulpt artifact that doesn't raise Attribute Errors...
            raise AttributeError
        return self.add_x_to_sym_table(key, value, self.class_table)

    def add_func_to_sym_table(self, ins_node, std_node):
        """
        Adds ins_node.name to the symbol table if it doesn't already exist, mapping it to a set of ins_node. Updates a
        second dictionary that maps ins_node to an std_node, and overwrites the current std_node since there should only
        be one mapping.

        Args:
            ins_node: instructor node or str representing a function name
            std_node: student node representing function

        Returns:
            int: number of conflicts generated

        """
        if not isinstance(std_node, CaitNode):
            raise TypeError
        if isinstance(ins_node, str):
            key = ins_node
        else:
            try:
                if ins_node.ast_name == "FunctionDef":
                    key = ins_node.astNode.name
                else:  # TODO: Little skulpt artifact that doesn't raise Attribute Errors...
                    key = ins_node._id
                    raise AttributeError
            except AttributeError:
                key = ins_node.astNode._id

        try:
            if std_node.ast_name == "FunctionDef":
                value = AstSymbol(std_node.astNode.name, std_node)
            else:  # TODO: Little skulpt artifact that doesn't raise Attribute Errors...
                raise AttributeError
        except AttributeError:
            node = std_node
            if type(node.astNode).__name__ != "Call":
                node = node.parent
                node._id = std_node._id
            value = AstSymbol(std_node._id, node)
        return self.add_x_to_sym_table(key, value, self.func_table)

    def add_var_to_sym_table(self, ins_node, std_node):
        """
        Adds ins_node._id to the symbol table if it doesn't already exist, mapping it to a set of ins_node. Updates a
        second dictionary that maps ins_node to an std_node, and overwrites the current std_node since there should only
        be one mapping.

        Args:
            ins_node: instructor node or str representing variable
            std_node: student node representing variable

        Returns:
            int: number of conflicts generated

        """
        if not isinstance(std_node, CaitNode):
            raise TypeError
        if isinstance(ins_node, str):
            key = ins_node
        else:
            key = ins_node.astNode._id
        value = AstSymbol(std_node.astNode._id, std_node)
        return self.add_x_to_sym_table(key, value, self.symbol_table)

    def add_exp_to_sym_table(self, ins_node, std_node):
        """
        Adds mapping of expression symbol to student node
        This function does NOT check for conflicts at the moment and probably should at some point.
        TODO: Check for conflicts
        Args:
            ins_node: Instructor node representing an expression
            std_node: student ast subtree corresponding to the symbol

        Returns:
            None
        """
        if not isinstance(std_node, CaitNode):
            raise TypeError
        self.exp_table[ins_node.astNode.id] = std_node

    def add_node_pairing(self, ins_node, std_node):
        """
        Adds a mapping of instructor ast node to a specific student ast node
        Args:
            ins_node: instructor pattern ast node
            std_node: student ast node

        Returns:
            None
        """
        if not isinstance(std_node, CaitNode):
            raise TypeError
        self.mappings[ins_node] = std_node

    def has_conflicts(self):
        """

        Returns:
            bool: True if number of conflicts is greater than 0
        """
        return len(self.conflict_keys) > 0

    def new_merged_map(self, other):
        """
        Returns a newly merged map consisting of this and other
        without modifying self.
        Args:
            other (AstMap): the other AstMap to be merged with

        Returns:
            AstMap: self modified by adding the contents of other
        """
        new_map = AstMap()
        new_map.merge_map_with(self)
        new_map.merge_map_with(other)
        return new_map

    def merge_map_with(self, other):
        """
        Returns a newly merged map consisting of this and other
        by modifying self
        Args:
            other (AstMap): the other AstMap to be merged with

        Returns:
            AstMap: self modified by adding the contents of other
        """
        if other is None:
            return

        if not isinstance(other, type(self)):
            raise TypeError

        # merge all mappings
        self.mappings.update(other.mappings)

        # merge all expressions
        self.exp_table.update(other.exp_table)

        # merge all symbols
        for key, value in other.symbol_table.items():
            for sub_value in value:
                self.add_var_to_sym_table(key, sub_value.astNode)

        # merge all functions
        for key, value in other.func_table.items():
            for sub_value in value:
                self.add_func_to_sym_table(str(key), sub_value.astNode)

        # merge all classes
        for key, value in other.class_table.items():
            for sub_value in value:
                self.add_class_to_sym_table(str(key), sub_value.astNode)

    @property
    def match_lineno(self):
        """
        Returns:
            int: the line number this match started on
        """
        values = [v.lineno for v in self.mappings.values()
                  if v.lineno is not None]
        if not values:
            return -1
        else:
            return min(values)

    def match_location(self, first=True):
        """
        Determines the first (or last) location of this match.

        Args:
            first (bool): Whether to use the first found match's location, or
                the last one.

        Returns:
            pedal.core.Location: The the first location of this match.
        """
        finder = min if first else max
        values = [v.lineno for v in self.mappings.values()
                  if v.lineno is not None]
        if not values:
            return Location(None)
        else:
            return Location(finder(values))

    def __getitem__(self, id_n):
        if id_n.startswith('__'):
            expression = self.exp_table[id_n]
            expression.map = self
            return expression
        else:
            if id_n in self.symbol_table:
                return self.symbol_table[id_n]
            elif id_n in self.func_table:
                return self.func_table[id_n]
            else:
                return self.class_table[id_n]

    def __contains__(self, id_n):
        if id_n.startswith('__'):
            return id_n in self.exp_table
        else:
            exists = id_n in self.symbol_table
            if exists:
                return exists
            else:
                return id_n in self.func_table

    def __len__(self):
        return len(self.mappings)

    def names(self):
        return {
            k: v.id for k, v in
            self.symbol_table.items()
        }



"""
Initialize and setup the Questions tool.
"""

from pedal.core.report import MAIN_REPORT, Report
from pedal.questions.constants import TOOL_NAME

import hashlib


def _name_hash(name: str) -> int:
    """ Turn a string name into a unique hashcode using MD5 """
    name = ",".join(str(x) for x in bytes(name, 'utf-8'))
    return hashlib.md5(name.encode('utf8')).digest()[0]


def reset(report=MAIN_REPORT):
    """
    Remove all settings for the Questions tool. Seed will be 0.

    Args:
        report: The report object

    Returns:
        This tool's data
    """
    report[TOOL_NAME] = {
        'seed': 0
    }
    return report[TOOL_NAME]


Report.register_tool(TOOL_NAME, reset)


def set_seed(seed_value, report=MAIN_REPORT):
    """
    Sets the seed that will be used in selecting questions.

    Args:
        seed_value (int or str or iterable[int]): The value to use when
            selecting questions, deterministically. If int, the same index
            will be used for all questions. If an iterable of ints, each
            one will serve as the index for the corresponding problem (throws
            an exception if the iterable isn't long enough). If a string,
            it will be hashed to a value (the hash is deterministic across
            platforms) that will be modulo'd to be in the right range for the
            pool. Presently, hashing generates values from [0, 256) so you
            need to limit your questions to 256.
        report (Report): The report object to store data and feedback in. If
            left None, defaults to the global MAIN_REPORT.
    """
    report['questions']['seed'] = seed_value



"""
Useful booleans to determine what version of Python we are using.
"""

import sys

IS_PYTHON_36 = sys.version_info[:2] == (3, 6)
IS_PYTHON_37 = sys.version_info[:2] == (3, 7)
IS_PYTHON_38 = sys.version_info[:2] == (3, 8)
IS_AT_LEAST_PYTHON_38 = sys.version_info[:2] >= (3, 8)
IS_AT_LEAST_PYTHON_39 = sys.version_info[:2] >= (3, 9)
IS_AT_LEAST_PYTHON_310 = sys.version_info[:2] >= (3, 10)
IS_AT_LEAST_PYTHON_311 = sys.version_info[:2] >= (3, 11)
IS_SKULPT = sys.platform == 'skulpt'



'''

TODO: This is not yet working!

**Sections** are a way to separate the pieces of a file such that the pieces do not
interfere with each other.

**Phases** are a way to chunk a collection of functions together. If one of these
functions fails, the other functions in the phase will continue to be evaluated.
However, that phase will still have failed. You can establish that one phase
comes before or after another phase; if a precondition phase fails, then the
subsequent phase will not run.

Example:
    Students are working on a text adventure game and have to implement a
    function named create_world(). The grading for portion of the assignment
    has three phases:
        'create_world_exists' which confirms that the function was defined
        'create_world_returns' which confirms that calling the function
            produces the right result.
        'create_world_complete' which confirms that the previous phase
            terminated in order to give some partial credit.
    
    Although the 'create_world_exists' phase is composed of one function, the
    'create_world_returns' phase is actually composed of several functions that
    check the components of the function.
    
    @phase('create_world_exists')
    
    @phase('create_world_returns', after='create_world_exists')
    
Phases are reset between sections.

**Subtests** are a way to combine multiple assertions into a single cohesive test.
TODO: Rename? And expand on.
with assert_group('test_add') as test_add_group:

if assert_unit_tests('function_name', FUNCTION_CASES):


'''
from pedal.assertions.feedbacks import AssertionBreak
from pedal.core.report import MAIN_REPORT
from pedal.assertions.setup import (_setup_assertions, _add_relationships, _add_phase)
from functools import wraps

def phase(phase_name, before=None, after=None):
    '''
    
    Args:
        phase_name (str): The name of the phase this function will belong to.
        before (list[str] or str): the name(s) of any phases that this phase
                                   should be before.
        after (list[str] or str): the name(s) of any phases that this phase
                                  should be after.
    '''
    _setup_assertions(MAIN_REPORT)
    def wrap(f):
        """

        Args:
            f:

        Returns:

        """
        @wraps(f)
        def _handle_entry(*args, **kwargs):
            old_exception_state = MAIN_REPORT['assertions']['exceptions']
            MAIN_REPORT['assertions']['exceptions'] = True
            value = f(*args, **kwargs)
            MAIN_REPORT['assertions']['exceptions'] = old_exception_state
            return value
        _add_phase(phase_name, _handle_entry)
        _add_relationships(phase_name, before)
        _add_relationships(after, phase_name)
        return _handle_entry
    return wrap


def stop_on_failure(f):
    """

    Args:
        f:

    Returns:

    """
    _setup_assertions(MAIN_REPORT)
    @wraps(f)
    def wrapped(*args, **kwargs):
        """

        Args:
            *args:
            **kwargs:

        Returns:

        """
        old_exception_state = MAIN_REPORT['assertions']['exceptions']
        MAIN_REPORT['assertions']['exceptions'] = True
        value = None
        try:
            value = f(*args, **kwargs)
        except AssertionBreak:
            pass
        MAIN_REPORT['assertions']['exceptions'] = old_exception_state
        return value
    return wrapped


def try_all(f):
    """

    Args:
        f:

    Returns:

    """
    _setup_assertions(MAIN_REPORT)
    @wraps(f)
    def wrapped(*args, **kwargs):
        """

        Args:
            *args:
            **kwargs:

        Returns:

        """
        old_exception_state = MAIN_REPORT['assertions']['exceptions']
        MAIN_REPORT['assertions']['exceptions'] = False
        value = f(*args, **kwargs)
        MAIN_REPORT['assertions']['exceptions'] = old_exception_state
        return value
    return wrapped


def precondition(function):
    """

    Args:
        function:
    """
    pass


def postcondition(function):
    """

    Args:
        function:
    """
    pass



"""
Simple data class for storing information about a location within source code.
"""
from dataclasses import dataclass

__all__ = ['Location']


@dataclass
class Location:
    """
    A class for storing information about a location in source code.

    Attributes:
        line (int): A line of source code.
        col (int, optional): A column within a line of source code.
            If missing, then defaults to the entire line.
        end_line (int, optional): The ending line of the source code region.
            Requires :py:attr:`line`.
        end_col (int, optional): The ending column of the source code region.
            Requires :py:attr:`col`.
        filename (str, optional): The filename that this location refers to.
            If missing, then defaults to the student's submission's main file.
    """

    def __init__(self, line, col=None, end_line=None, end_col=None,
                 filename=None):
        self.line = line
        self.col = col
        self.end_line = end_line
        self.end_col = end_col
        self.filename = filename

    def __str__(self):
        return f"<Location({self.line}, {self.col}, {self.filename!r})>"

    def __repr__(self):
        return str(self)

    @classmethod
    def from_ast(cls, node):
        """
        Creates a new Location object from the AST node. Should work
        for both built-in AST nodes and CaitNodes.

        Args:
            node (Node):

        Returns:
            Location
        """
        return Location(node.lineno, col=node.col_offset)

    def to_json(self):
        """
        Creates a JSON version of this object, with all the fields.

        Returns:
            Dict[str,Any]: The JSON version of this location information.
        """
        return {
            "line": self.line,
            "col": self.col,
            "end_line": self.end_line,
            "end_col": self.end_col,
            "filename": self.filename
        }



from pedal.core.report import MAIN_REPORT, Report
from pedal.sandbox.constants import TOOL_NAME
from pedal.sandbox.sandbox import Sandbox
from pedal.sandbox.commands import *


def reset(report=MAIN_REPORT):
    """
    Resets (or initializes) the Sandbox instance for this report.
    Completely destroys the existing Sandbox instance attached to this report;
    if you want to just clear out its data, then instead use
    :py:func:`pedal.sandbox.commands.clear_sandbox`.

    Args:
        report: The report to reset the Sandbox for.
    """
    report[TOOL_NAME] = {
        'sandbox': Sandbox(report=report)
    }


Report.register_tool(TOOL_NAME, reset)



import os
import sys
from textwrap import indent, dedent


def parse_out_file(path):
    """
    Parses an out file, which is based on ``ini`` files. They have pretty
    basic syntax. The difference from what configparser is that indentation
    is preserved in multiline values. We don't support comments and a number
    of other features, but who needs 'em.

    Args:
        path (str): The path to the inifile that will be read.

    Returns:
        dict[str, dict[str, str]]: The constructed out sections=>{fields=>data}
    """
    sections = {'default': {}}
    current_section = 'default'
    current_key = None
    multiline = False
    with open(path) as outfile:
        for line in outfile:
            if line.startswith('['):
                multiline = False
                current_section = line.strip()[1:-1]
                sections[current_section] = {}
            elif line.startswith('    '):
                sections[current_section][current_key] += line[4:]
            elif line.strip():
                multiline = False
                parts = line.split(":", maxsplit=1)
                current_key = parts[0].rstrip()
                sections[current_section][current_key] = ""
                if len(parts) == 1 or not parts[1].lstrip():
                    multiline = True
                else:
                    sections[current_section][current_key] += parts[1].strip()
            elif multiline:
                sections[current_section][current_key] += "\n"
    return sections


def generate_out_file(outfile, sections):
    for section_name, section in sections.items():
        outfile.write(f"[{section_name}]\n")
        for field, value in section.items():
            outfile.write(f"{field}:")
            value = str(value)
            if '\n' in value:
                outfile.write("\n")
                for line in value.split("\n"):
                    outfile.write(f"    {line}\n")
            else:
                outfile.write(f" {value}\n")


def _make_final(environment):
    """ Determine the final environment's section """
    if environment is None:
        return 'final'
    else:
        return f"{environment}.final"


class ReportVerifier:
    """ Helper class for reading and processing a '.out' file. """
    def __init__(self, path, environment):
        self.outfile = parse_out_file(path)
        self.environment = environment

    def get_final(self):
        """ Get the expected final feedback. """
        final_environment = _make_final(self.environment)
        if final_environment in self.outfile:
            return self.outfile[final_environment].items()
        elif 'final' in self.outfile:
            return self.outfile['final'].items()
        return {}


def generate_report_out(path, environment, report):
    result = {}
    # Allow graceful appending
    if path is not None and os.path.exists(path):
        result = parse_out_file(path)
    final_environment = _make_final(environment)
    try:
        fields = report.result.to_json()
    except Exception as e:
        raise report
    # TODO: Hack, just remove the data field for simplicity - fix this later
    fields.pop('data')
    fields.pop('hide_correctness')
    result[final_environment] = fields
    # TODO: Support ``forbid`` and ``available`` fields for finer grained
    #   output checking.
    if path is not None:
        with open(path, 'w') as outfile:
            generate_out_file(outfile, result)
    else:
        generate_out_file(sys.stdout, result)
        return result



"""
File that holds the the Report class and the global MAIN_REPORT.

Note that you can make other Reports, but that doesn't actually seem to be
useful very often. Usually you want to just rely on the global MAIN_REPORT.
"""

__all__ = ['Report', 'MAIN_REPORT']
import logging
import random

from pedal.core.errors import PedalToolNotRegistered, PedalToolAlreadyRegistered
from pedal.core.feedback_category import FeedbackCategory


# TODO: Mechanism for checking whether a piece of feedback is in the report
from pedal.core.formatting import Formatter
from pedal.core.tool import ToolRegistration

log = logging.getLogger(__name__)


class Report:
    """
    A class for storing Feedback generated by Tools, along with any auxiliary
    data that the Tool might want to provide for other tools.

    Attributes:
        submission (:py:class:`~pedal.core.submission.Submission`): The
            contextualized submission information.
        feedback (list[:py:class:`~pedal.core.feedback.Feedback`]): The raw
            feedback generated for this Report so far.
        suppressions (list[tuple[str, str]]): The categories and labels that
            have been suppressed so far.
        hiddens (set[str]): The parts of the final response that should be
            hidden. This can globally hide the 'correct', 'score', etc.
        group (int or str): The label for the current group. Feedback given
            by a Tool will automatically receive the current `group`. This
            is used by the Source tool, for example, in order to group feedback
            by sections and the :py:func:`pedal.assertions.commands.unit_test`
            function to combine results.
        group_names (dict[group, str]): A printable, student-facing name for the
            group. When a group needs to be rendered out to the user, this
            will override whatever label was going to be presented instead.
        hooks (dict[str, list[callable]): A dictionary mapping events to
            a list of callable functions. Tools can register functions on
            hooks to have them executed when the event is triggered by another
            tool. For example, the Assertions tool has hooks on the Source tool
            to trigger assertion resolutions before advancing to next sections.
        _tool_data (dict[str, Any]): Maps tool names to their data. The
                                       namespace for a tool can be used to
                                       store whatever they want, but will
                                       probably be in a dictionary itself.
        resolves (list[Any]): The result of having previously called a
            resolver. This allows you to check if a report has previously
            been resolved, or do something with that data.
        result (FinalFeedback): The FinalFeedback (distinct from a Feedback) that was
            generated as a result of resolving this Report, or None if the Report is
            not yet resolved.
    """
    #: dict[str, dict]: The
    #: tools registered for this report, available via their names.
    TOOLS = {}

    def __init__(self):
        """
        Creates a new Report instance.
        """
        self._tool_data = {}
        self.feedback = []
        self.ignored_feedback = []
        self.suppressions = {}
        self.suppressed_labels = {}
        self.hiddens = set()
        self.groups = []
        self.group = None
        self.group_names = {}
        self.hooks = {}
        self.class_hooks = {}
        self.submission = None
        self.format = Formatter()
        self.result = None
        self.resolves = []
        self.pools = []
        self.chosen_pool = None
        self.overridden_feedbacks = set()
        log.debug("New Pedal Report created.")

    def clear(self):
        """
        Resets the entire report back to its starting form,
        including deleting any attached submissions, tool data, and feedbacks.
        It will also reset any overridden fields of feedback classes.
        However, it will not affect class hooks.
        """
        self.feedback.clear()
        self.ignored_feedback.clear()
        self.suppressions.clear()
        self.suppressed_labels.clear()
        self.hiddens.clear()
        self._tool_data.clear()
        self.group = None
        self.groups.clear()
        self.group_names.clear()
        self.hooks.clear()
        self.submission = None
        self.result = None
        self.resolves.clear()
        self.format = Formatter()
        self.clear_overridden_feedback()

    def full_clear(self):
        """ This totally resets the report, including any class hooks. """
        self.clear()
        self.class_hooks.clear()

    def clear_overridden_feedback(self):
        for feedback in self.overridden_feedbacks:
            feedback._restore_overrides()
        self.overridden_feedbacks.clear()

    def override_feedback(self, feedback_class):
        self.overridden_feedbacks.add(feedback_class)

    def set_pools(self, pools):
        if isinstance(pools, int):
            pools = list('ABCDEFGHIJKLMNOPQRSTUVWXYZ'[:pools])
        self.pools = pools

    def contextualize(self, submission):
        """
        Attach the given submission to this report.

        Args:
            submission (:py:class:`pedal.core.submission.Submission`): The
                submission to attach to this report.
        """
        self.submission = submission

    def hide_correctness(self):
        """
        Suppress the RESULT category entirely, so that the report doesn't
        indicate whether or not the submission was correct.
        TODO: Make this just a regular command.
        """
        self.hiddens.add('correct')
        self.hiddens.add('score')

    def add_feedback(self, feedback):
        """
        Attaches the given feedback object to this report.

        Args:
            feedback (:py:class:`~pedal.core.feedback.Feedback`): The feedback
                object to attach.

        Returns:
            :py:class:`~pedal.core.feedback.Feedback`: The attached feedback.
        """
        self.feedback.append(feedback)
        if not isinstance(feedback.parent, (int, str)) and feedback.parent is not None:
            feedback.parent._get_child_feedback(feedback, True)
        self.execute_hooks('pedal.report', 'add_feedback', (feedback,))
        return feedback

    def add_ignored_feedback(self, feedback):
        """
        Attaches the given feedback object to this report, but only in the
        ignored list. That means it should not be considered by the Resolver,
        since its condition did not apply to the code. Some Resolvers like
        to know about feedback that was not reached.

        Args:
            feedback (:py:class:`~pedal.core.feedback.Feedback`): The feedback
                object to attach.

        Returns:
            :py:class:`~pedal.core.feedback.Feedback`: The attached feedback.
        """
        self.ignored_feedback.append(feedback)
        if feedback.parent is not None:
            feedback.parent._get_child_feedback(feedback, False)
        return feedback

    def suppress(self, category=None, label=True, fields=None):
        """
        Suggest that an entire category or label within a category ignored by
        the resolver.
        TODO: Currently, only global suppression is supported.

        Args:
            category (str): The category of feedback to suppress.
            label (bool or str): A specific label to match against and suppress.
            fields (dict of key/values): The fields that will be matched exactly to
                suppress.
        """
        if fields is None:
            fields = {}
        if category is None:
            if label not in self.suppressed_labels:
                self.suppressed_labels[label] = []
            self.suppressed_labels[label].append(fields)
        else:
            category = category.lower()
            if isinstance(label, str):
                label = label.lower()
            if category in FeedbackCategory.ALIASES:
                category = FeedbackCategory.ALIASES[category]
            if category not in self.suppressions:
                self.suppressions[category] = {}
            if label not in self.suppressions[category]:
                self.suppressions[category][label] = []
            self.suppressions[category][label].append(fields)

    def add_hook(self, event, function):
        """
        Register the `function` to be executed when the given `event` is
        triggered.
        
        Args:
            event (str): An event name. Multiple functions can be triggered for
                the same `event`. The format is as follows: `"namespace.function.extra"`
                The `".extra"` component is optional to add further nuance, but
                the general idea is that you are referring to functions that,
                when called, should trigger other functions to be called first.
                The namespace is typically a tool or module.
            function (callable): A callable function. This function should
                accept a keyword parameter named `report`; this report will be passed
                as as that argument.
        """
        if event not in self.hooks:
            self.hooks[event] = []
        self.hooks[event].append(function)

    @classmethod
    def add_class_hook(cls, event, function):
        """ Similar to ``add_hook``, except attaches them to the class, so
        they will be executed for ALL report subclasses. """
        if event not in cls.class_hooks:
            cls.class_hooks[event] = []
        cls.class_hooks[event].append(function)

    def execute_hooks(self, tool, event_name, arguments=None, keyword_arguments=None):
        """
        Trigger the functions for all of the associated hooks.
        Hooks will be called with this report as a keyword `report` argument.

        Args:
            tool (str): The name of the tool, to namespace events by.
            event_name (str): The event name (separate words with periods).
            arguments (tuple[any]): The arguments to be passed to the callback function.
            keyword_arguments (dict[str, any]): The keyword arguments to be passed to the callback funciton.
        """
        if arguments is None:
            arguments = tuple()
        if keyword_arguments is None:
            keyword_arguments = {}
        event = tool + '.' + event_name
        if event in self.class_hooks:
            for function in self.class_hooks[event]:
                function(report=self, *arguments, **keyword_arguments)
        if event in self.hooks:
            for function in self.hooks[event]:
                function(report=self, *arguments, **keyword_arguments)

    def __getitem__(self, tool_name):
        """
        Support retrieving a tool's data from the report using square bracket
        syntax. So, for example, you can do `MAIN_REPORT['tifa']` and get its
        data dictionary. If the tool has been registered, but not initialized
        for this report, then the tool will be
        :py:method:`pedal.core.tool.reset` first. Otherwise, throws an error
        that the tool does not exist.

        Args:
            tool_name (str): The formal name of the tool, most likely specified
                in its `constants.py` file.

        Returns:
            dict: The data associated with that tool.
        """
        if tool_name not in self._tool_data:
            if tool_name not in self.TOOLS:
                raise PedalToolNotRegistered(tool_name, list(self.TOOLS.keys()))
            self.TOOLS[tool_name].reset(report=self)
        return self._tool_data[tool_name]

    def __setitem__(self, tool_name, value):
        """
        Update the tool's current data. Should largely not be used by anyone.
        In fact, this could seriously damage the relationships between tools.

        Args:
            tool_name (str): The name of the tool.
            value (dict): The new data to set as this tool's namespace.
        """
        self._tool_data[tool_name] = value

    def __contains__(self, tool_name):
        """
        Determine if the given `tool_name` is available through this report.
        Args:
            tool_name (str): The name of a tool.

        Returns:
            bool: Whether the tool is available.
        """
        return tool_name in self._tool_data

    def set_formatter(self, formatter):
        """
        Update the formatter with the new option.

        Args:
            formatter (:py:class:`pedal.core.formatting.Formatter`): The new
                formatter to use.
        """
        self.format = formatter

    @classmethod
    def register_tool(cls, tool_name: str, reset_function):
        """
        Identifies that the given Tool should be made available.
        Args:
            tool_name: A unique string identifying this tool.
            reset_function: The function to call to reset the Tool.

        Returns:

        """
        if tool_name in cls.TOOLS:
            raise PedalToolAlreadyRegistered(tool_name)
        cls.TOOLS[tool_name] = ToolRegistration(tool_name, reset_function)

    def get_current_group(self):
        if self.groups:
            return self.groups[-1]
        else:
            return None

    def start_group(self, group):
        self.groups.append(group)

    def stop_group(self, group):
        """
        TODO: Should this prematurely end other groups? If so, do they get a
            callback event to do any wrapup?
        """
        if self.groups:
            self.groups.remove(group)

    def finalize_pools(self):
        if self.pools:
            self.chosen_pool = random.choice(self.pools)

    def finalize_feedbacks(self):
        self.finalize_pools()
        for feedback in self.feedback + self.ignored_feedback:
            feedback._finalize()

#: The global Report object. Meant to be used as a default singleton
#: for any tool, so that instructors do not have to create their own Report.
#: Of course, all APIs are expected to work with a given Report, and only
#: default to this Report when no others are given.
#: Ideally, the average instructor will never know this exists.
MAIN_REPORT = Report()

# TODO: Give a mechanism for "freezing" a report that you can keep around.



"""
The assertions module contains classic unittest-style assert statements.
"""

from pedal.assertions.setup import _setup_assertions, resolve_all
from pedal.assertions.constants import TOOL_NAME
from pedal.core.report import Report, MAIN_REPORT
from pedal.assertions.commands import *
from pedal.assertions.positive import close_output, correct_output



def reset(report=MAIN_REPORT):
    """
    Resets (or initializes) the information about assertions.

    Args:
        report:
    """
    report[TOOL_NAME] = {
        'failures': 0,
        'exceptions': False
    }


Report.register_tool(TOOL_NAME, reset)



from textwrap import dedent

from pedal.assertions.feedbacks import assert_group
from pedal.assertions.runtime import *
from pedal.sandbox.commands import call, start_trace, get_sandbox, evaluate, CommandBlock, run
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE
from pedal.sandbox.library.drafter import MockDrafter
import unittest


@unittest.skip
class TestAssertions(ExecutionTestCase):

    def test_basic_mocking(self):
        with Execution(dedent("""
        from drafter import *
        from dataclasses import dataclass
        @dataclass
        class State:
            pass
        @route
        def index(state: State) -> Page:
            return Page(state, ["Hello World!"])
        start_server()
        """)) as e:
            pass # assert_route("index")
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_microbit_displayed_simple_2d_list_fails(self):
        with Execution(dedent("""
        from microbit.display import show
        # A 3x3 box with the bottom-left corner and center missing
        show([[0, 0, 0, 0, 0], [0, 9, 9, 9, 0], [0, 9, 0, 9, 0], [0, 0, 9, 9, 0], [0, 0, 0, 0, 0]])
        """)) as e:
            assert_microbit_displayed("     "
                                      " ██  "
                                      " █ █ "
                                      " ███ "
                                      "     ", report_differences=True)
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, """Image Not Displayed on Microbit
The expected image was not displayed on the Microbit.
The closest actual image shown had 2 differences.
The differences were at the following positions:
    3, 1: 9 instead of 0
    1, 3: 0 instead of 9""")

    def test_assert_microbit_displaying_2d_list_passes(self):
        with Execution(dedent("""
        from microbit.display import show
        # A 3x3 box with the bottom-left corner and center missing
        show([[0, 0, 0, 0, 0], [0, 9, 9, 9, 0], [0, 9, 0, 9, 0], [0, 0, 9, 9, 0], [0, 0, 0, 0, 0]])
        """)) as e:
            assert_microbit_displaying("     "
                                       " ███ "
                                       " █ █ "
                                       "  ██ "
                                       "     ", True)
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_microbit_displaying_2d_list_fails_current(self):
        with Execution(dedent("""
        from microbit.display import show
        # A 3x3 box with the bottom-left corner and center missing
        show([[0, 0, 0, 0, 0], [0, 9, 9, 9, 0], [0, 9, 0, 9, 0], [0, 0, 9, 9, 0], [0, 0, 0, 0, 0]])
        show([[0, 0, 0, 0, 0], [0, 9, 9, 9, 0], [0, 9, 9, 9, 0], [0, 0, 9, 9, 0], [0, 0, 0, 0, 0]])
        """)) as e:
            assert_microbit_displaying("     "
                                       " ███ "
                                       " █ █ "
                                       "  ██ "
                                       "     ", True)
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, """Image Not Displaying on Microbit
The expected image is not currently displaying on the Microbit.
The image actually shown had 1 differences.
The differences were at the following positions:
    2, 2: 9 instead of 0""")

    def test_assert_microbit_displayed_animation_touches_corners(self):
        with Execution(dedent("""
        from microbit.display import show
        # A 3x3 box with the bottom-left corner and center missing
        show([[0, 0, 0, 0, 9], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])
        show([[9, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])
        show([[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [9, 0, 0, 0, 0]])
        show([[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 9]])
        """)) as e:
            assert_microbit_displayed("    █:     :     :     :     ",
                                      "█    :     :     :     :     ",
                                      "     :     :     :     :█    ",
                                      "     :     :     :     :    █", report_differences=True)
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_microbit_displayed_animation_touches_corners_fails(self):
        with Execution(dedent("""
        from microbit import display

        show = display.show
        # A 3x3 box with the bottom-left corner and center missing
        show([[0, 0, 0, 0, 9], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])
        show([[9, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])
        show([[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [9, 0, 0, 0, 0]])
        show([[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 9]])
        """)) as e:
            assert_microbit_displayed("    █:     :     :     :     ",
                                      "█    :     :     :     :     ",
                                      "     :     :     :     :     ",
                                      "     :     :     :     :    █", report_differences=True)
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, """Image Not Displayed on Microbit
The expected image was not displayed on the Microbit.
The closest actual image shown had 1 differences.
The differences were at the following positions:
    4, 0: 0 instead of 9""")

    def test_assert_microbit_displayed_animation_touches_corners_loop(self):
        with Execution(dedent("""
        from microbit import display
        # A 3x3 box with the bottom-left corner and center missing
        STARTS = [(4, 0), (0, 0), (0, 4), (4, 4)]
        DIRECTIONS = [(-1, 0), (0, 1), (1, 0), (0, -1)]
        display.clear()
        previous_x, previous_y = (0, 0)
        for (start_x, start_y), (dx, dy) in zip(STARTS, DIRECTIONS):
            for i in range(0, 5):
                display.set_pixel(previous_x, previous_y, 0)
                previous_x, previous_y = start_x + dx*i, start_y + dy*i
                display.set_pixel(previous_x, previous_y, 9)
        """)) as e:
            assert_microbit_displayed("    █:     :     :     :     ",
                                      "   █ :     :     :     :     ",
                                      " █   :     :     :     :     ",
                                      "█    :     :     :     :     ",
                                      "     :     :     :     :█    ",
                                      "     :     :     :     :    █", report_differences=True)
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, SUCCESS_MESSAGE)


"""
Representation of a student's submission to pedal. Almost certainly contains their code,
but may also contain other metadata.


TODO: Normalize the concept of evaluations ("<stdin>" or "evaluations").
get_program(filename='<stdin>') => submission.files['<stdin>']
get_evaluation()

"""

__all__ = ['Submission']

from pedal.utilities.files import get_extension

DEFAULT_EXTENSION = "py"


def python_submission_parser(code: str, filename: str) -> (str, str):
    return code, filename


def create_ipynb_submission_parser(metadata_filter=None):
    def ipynb_submission_parser(code: str, filename: str) -> (str, str):
        # TODO: What happens if we have a syntax error in the file?
        # TODO: Consider making a fallback JSON parser, since the structure is known.
        try:
            import nbformat
        except ImportError:
            raise ImportError("You have attempted to read an IPYNB file, but did not install `nbformat`. Is Jupyter installed?")
        nb = nbformat.reads(code, as_version=nbformat.current_nbformat)
        result = []
        for cell in nb.cells:
            if metadata_filter is not None and not metadata_filter(cell, nb.cells):
                continue
            if cell.cell_type == 'markdown':
                result.append(f'"""{cell.source}"""')
            elif cell.cell_type == 'code':
                result.append(cell.source)
        return "\n".join(result), filename
    return ipynb_submission_parser


class Submission:
    """
    Simple class for holding information about the student's submission.

    Examples:
        A very simple example of creating a Submission with a single file::

            >>> Submission({'answer.py': "a = 0"})

    Attributes:
        files (`dict` mapping `str` to `str`): Dictionary of filenames mapped to their contents, emulating
            a file system.
        main_file (str): The entry point file that will be considered the main file.
        main_code (str): The actual code to run; if None, then defaults to the code of the main file.
            Useful for tools that want to change the currently active code (e.g., Source's sections) or
            run additional commands (e.g., Sandboxes' call).
        user (dict): Additional information about the user.
        assignment (dict): Additional information about the assignment.
        course (dict): Additional information about the course.
        execution (dict): Additional information about the results of executing the students' code.
    """

    PARSERS = {
        "py": python_submission_parser,
        "ipynb": create_ipynb_submission_parser()
    }

    def __init__(self, files=None, main_file='answer.py', main_code=None,
                 user=None, assignment=None, course=None, execution=None,
                 instructor_file='instructor_tests.py', load_error=None):
        if files is None:
            files = {}
        self.files = files
        self.main_file = main_file
        self.load_error = load_error
        if main_code is not None:
            self.main_code = main_code
        if not self.main_code:
            self.main_code = ""
        if self.main_file not in self.files:
            self.main_code, self.main_file = self._preprocess_file(self.main_code, self.main_file)
            self.files[self.main_file] = self.main_code
        else:
            self.files[self.main_file], self.main_file = self._preprocess_file(self.files[self.main_file], self.main_file)
        # TODO: Non-main files are not going to be pre-processed; is that a problem for anyone?
        self._original_main_code = self.main_code
        self.user = user
        self.assignment = assignment
        self.course = course
        self.execution = execution
        self.instructor_file = instructor_file
        self.line_offsets = {}
        self._lines_cache = {}

    def get_lines(self, filename=None):
        """
        Retrieves the lines of code from this submission.

        Returns:
            list[str]: The lines of code for this submission.
        """
        if filename is None:
            code = self.main_code
        else:
            code = self.files[filename]
        if code not in self._lines_cache:
            self._lines_cache[code] = code.split("\n")
        return self._lines_cache[code]

    def get_files_lines(self):
        """ Retrieves a dictionary of lists of strings representing the files'
        lines of code. """
        return {filename: self.get_lines(filename) for filename in self.files}

    def replace_main(self, code: str, file: str = None):
        """
        Substitutes the current main code and filename with the given arguments.
        Args:
            code (str): The new code to substitute in.
            file (str): An optional filename to use.
        """
        self.main_code = code
        if file is not None:
            self.main_file = file
        code, self.main_file = self._preprocess_file(code, self.main_file)
        self.files[self.main_file] = code

    def _preprocess_file(self, code, filename):
        """
        Give the system a chance to potentially preprocess the given file, based on its extension.
        For example, to turn an IPYNB file into more conventional python code.

        Args:
            code:
            filename:

        Returns:

        """
        extension = get_extension(filename, DEFAULT_EXTENSION)
        parser = self.PARSERS.get(extension, python_submission_parser)
        return parser(code, filename)

    @property
    def main_code(self):
        return self.files.get(self.main_file, "")

    @main_code.setter
    def main_code(self, code):
        self.files[self.main_file] = code

    def set_line_offset(self, lineno, filename=None):
        """ Sets the line offset for the given filename. Defaults to main
        file."""
        if filename is None:
            filename = self.main_file
        self.line_offsets[filename] = lineno

    def clear_line_offsets(self):
        self.line_offsets.clear()

    def to_json(self):
        return dict(
            user=self.user,
            assignment=self.assignment,
            course=self.course,
            execution=self.execution,
            files=self.files.copy()
        )





"""
A pretty-printing dump function for the ast module.  The code was copied from
the ast.dump function and modified slightly to pretty-print.

Alex Leone (acleone ~AT~ gmail.com), 2010-01-30

From https://alexleone.blogspot.co.uk/2010/01/python-ast-pretty-printer.html
"""

from ast import AST, iter_fields, parse


def dump(node, annotate_fields=True, include_attributes=False, indent='  '):
    """
    Return a formatted dump of the tree in *node*.  This is mainly useful for
    debugging purposes.  The returned string will show the names and the values
    for fields.  This makes the code impossible to evaluate, so if evaluation is
    wanted *annotate_fields* must be set to False.  Attributes such as line
    numbers and column offsets are not dumped by default.  If this is wanted,
    *include_attributes* can be set to True.
    """

    def _format(_node, level=0):
        if isinstance(_node, AST):
            fields = [(a, _format(b, level)) for a, b in iter_fields(_node)]
            if include_attributes and _node._attributes:
                fields.extend([(a, _format(getattr(_node, a), level))
                               for a in _node._attributes])
            return ''.join([
                _node.__class__.__name__,
                '(',
                ', '.join(('%s=%s' % field for field in fields)
                          if annotate_fields else
                          (b for a, b in fields)),
                ')'])
        elif isinstance(_node, list):
            lines = ['[']
            lines.extend((indent * (level + 2) + _format(x, level + 2) + ','
                          for x in _node))
            if len(lines) > 1:
                lines.append(indent * (level + 1) + ']')
            else:
                lines[-1] += ']'
            return '\n'.join(lines)
        return repr(_node)

    if not isinstance(node, AST):
        raise TypeError('expected AST, got %r' % node.__class__.__name__)
    return _format(node)


def parseprint(code, filename="<string>", mode="exec", **kwargs):
    """Parse some code from a string and pretty-print it."""
    node = parse(code, mode=mode)  # An ode to the code
    print(dump(node, **kwargs))


# Short name: pdp = parse, dump, print
pdp = parseprint



"""
Holds the SandboxResult proxy that emulates other classes as perfectly
as possible.
"""

# This module provides a version of ``len`` that handles SandboxResult
#   correctly, requiring us to hold a reference to the original ``len``.
import math

_original_len = len


class SandboxResult:
    """
    Proxy class for wrapping results from executing student code. Attempts
    to perfectly emulate the underlying data value, so that users will never
    realize they have a proxy. The advantage is that special information is
    available in the corresponding Sandbox about this result that can give
    more context.
    
    Attributes:
        value (any): The actual data stored in this class that we are proxying.
            If the underlying proxy object has a field called `value`, then
            you can use either `_actual_value` to access the proxied object.
        _actual_context_id (int): The call that was used to generate this result.
        _actual_sandbox (Sandbox): The sandbox that was used to generate this
            result. If None, then the sandbox was lost.
    
    """
    ASSIGNABLE_ATTRS = ['value', '_actual_context_id', '_actual_sandbox',
                        '_clone_this_result']

    def __init__(self, value, context_id=None, sandbox=None):
        """
        Args:
            value (any): Literally any type of data.
            context_id (int): The unique call ID that generated this result. If
                None, then the SandboxResult was generated by manipulating an earlier
                result.
                TODO: We could actually remember the operations applied to this
                    instance and use them to reconstruct the transformations...
            sandbox (Sandbox): The sandbox that was used to generate this
                result. If None, then the sandbox was lost.
        """
        self.value = value
        self._actual_context_id = context_id
        self._actual_sandbox = sandbox

    def __getattribute__(self, name):
        """
        Get the attribute with the given `name`. This allows us to pass
        most attributes along to the underlying `value`, while still
        maintaining access to the proxy's attributes.
        """
        v = object.__getattribute__(self, "value")
        if name == "__class__":
            return v.__class__
        elif name == "__actual_class__":
            return object.__getattribute__(self, "__class__")
        elif name == "_actual_value":
            return v
        elif name in SandboxResult.ASSIGNABLE_ATTRS:
            return object.__getattribute__(self, name)
        elif name == "value" and not hasattr(v, "value"):
            return v
        else:
            return SandboxResult(object.__getattribute__(v, name),
                                 object.__getattribute__(self, "_actual_context_id"),
                                 object.__getattribute__(self, "_actual_sandbox"))

    def __setattr__(self, name, value):
        if name in SandboxResult.ASSIGNABLE_ATTRS:
            object.__setattr__(self, name, value)
        else:
            setattr(self.value, name, value)

    def __delattr__(self, name):
        if name in SandboxResult.ASSIGNABLE_ATTRS:
            object.__delattr__(self, name)
        else:
            delattr(self.value, name)

    def _clone_this_result(self, new_value):
        """
        Create a new SandboxResult based on this current one. Copies over the
        `context_id` and `sandbox`.
        
        Args:
            new_value (any): The new value to be proxying.
        Returns:
            SandboxResult
        """
        return SandboxResult(new_value,
                             context_id=self._actual_context_id,
                             sandbox=self._actual_sandbox)

    def __repr__(self):
        """
        Returns the representation of the proxied object.
        
        Returns:
            str: The `repr` of the proxied object.
        """
        return repr(self.value)

    def __str__(self):
        """
        Returns the string representation of the proxied object.
        
        Returns:
            str: The `str` of the proxied object.
        """
        return str(self.value)

    def __bytes__(self):
        return bytes(self.value)

    def __format__(self, format_spec):
        return format(self.value, format_spec)

    def __call__(self, *args):
        """
        Returns the result of calling the proxied object with the args.
        
        Returns:
            SandboxResult: A proxy of the Sandbox object.
        """
        return self._clone_this_result(self.value(*args))

    def __hash__(self):
        return hash(self.value)

    def __bool__(self):
        return bool(self.value)

    def __dir__(self):
        return dir(self.value)

    def __instancecheck__(self, instance):
        return isinstance(self.value, instance)

    def __subclasscheck__(self, subclass):
        return issubclass(self.value, subclass)

    def __len__(self):
        """
        Fun fact: cpython DEMANDS that __len__ return an integer. Not something
        that looks like an integer, but a true, honest-to-god integer that
        can fit into a slot.
        https://stackoverflow.com/questions/42521449/how-does-python-ensure-the-return-value-of-len-is-an-integer-when-len-is-cal
        """
        return _original_len(self.value)

    def __getitem__(self, key):
        return self._clone_this_result(self.value[key])

    def __setitem__(self, key, value):
        self.value[key] = value

    def __delitem__(self, key):
        del self.value[key]

    def __missing__(self, key):
        return self.value.__missing__(key)

    def __iter__(self):
        return iter(self.value)

    def __reversed__(self):
        return reversed(self.value)

    def __contains__(self, item):
        return self.value.__contains__(item)

    def __eq__(self, other):
        """
        Test if the proxied object is equal to the given `other`.
        
        Args:
            other (any): The other object.
        
        Returns:
            bool or any: Returns whatever the proxy object's __eq__ returns.
        """
        if isinstance(other, SandboxResult):
            return self.value == other.value
        return self.value == other

    def __lt__(self, other):
        if isinstance(other, SandboxResult):
            return self.value < other.value
        return self.value < other

    def __le__(self, other):
        if isinstance(other, SandboxResult):
            return self.value <= other.value
        return self.value <= other

    def __gt__(self, other):
        if isinstance(other, SandboxResult):
            return self.value > other.value
        return self.value > other

    def __ge__(self, other):
        if isinstance(other, SandboxResult):
            return self.value >= other.value
        return self.value >= other

    def __ne__(self, other):
        if isinstance(other, SandboxResult):
            return self.value != other.value
        return self.value != other

    # Numeric Operations

    def __add__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value + other.value)
        return self._clone_this_result(self.value + other)

    def __sub__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value - other.value)
        return self._clone_this_result(self.value - other)

    def __mul__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value * other.value)
        return self._clone_this_result(self.value * other)

    def __matmul__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__matmul__(other.value))
        return self._clone_this_result(self.value.__matmul__(other))

    def __truediv__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__truediv__(other.value))
        return self._clone_this_result(self.value.__truediv__(other))

    def __floordiv__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__floordiv__(other.value))
        return self._clone_this_result(self.value.__floordiv__(other))

    def __mod__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__mod__(other.value))
        return self._clone_this_result(self.value.__mod__(other))

    def __divmod__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__divmod__(other.value))
        return self._clone_this_result(self.value.__divmod__(other))

    def __pow__(self, other, *modulo):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__pow__(other.value, *modulo))
        return self._clone_this_result(self.value.__pow__(other, *modulo))

    def __lshift__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__lshift__(other.value))
        return self._clone_this_result(self.value.__lshift__(other))

    def __rshift__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__rshift__(other.value))
        return self._clone_this_result(self.value.__rshift__(other))

    def __and__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__and__(other.value))
        return self._clone_this_result(self.value.__and__(other))

    def __xor__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__xor__(other.value))
        return self._clone_this_result(self.value.__xor__(other))

    def __or__(self, other):
        if isinstance(other, SandboxResult):
            return self._clone_this_result(self.value.__or__(other.value))
        return self._clone_this_result(self.value.__or__(other))

    def __radd__(self, other):
        if isinstance(self.value, str):
            return self._clone_this_result(self.value.__add__(other))
        return self._clone_this_result(self.value.__radd__(other))

    def __rsub__(self, other):
        return self._clone_this_result(self.value.__rsub__(other))

    def __rmul__(self, other):
        return self._clone_this_result(self.value.__rmul__(other))

    def __rmatmul__(self, other):
        return self._clone_this_result(self.value.__rmatmul__(other))

    def __rtruediv__(self, other):
        return self._clone_this_result(self.value.__rtruediv__(other))

    def __rfloordiv__(self, other):
        return self._clone_this_result(self.value.__rfloordiv__(other))

    def __rmod__(self, other):
        return self._clone_this_result(self.value.__rmod__(other))

    def __rdivmod__(self, other):
        return self._clone_this_result(self.value.__rdivmod__(other))

    def __rpow__(self, other):
        return self._clone_this_result(self.value.__rpow__(other))

    def __rlshift__(self, other):
        return self._clone_this_result(self.value.__rlshift__(other))

    def __rand__(self, other):
        return self._clone_this_result(self.value.__rand__(other))

    def __rxor__(self, other):
        return self._clone_this_result(self.value.__rxor__(other))

    def __ror__(self, other):
        return self._clone_this_result(self.value.__ror__(other))

    # TODO: __iadd__ and other in-place assignment operators?

    def __neg__(self):
        return self._clone_this_result(self.value.__neg__())

    def __pos__(self):
        return self._clone_this_result(self.value.__pos__())

    def __abs__(self):
        return self._clone_this_result(self.value.__abs__())

    def __invert__(self):
        return self._clone_this_result(self.value.__invert__())

    def __complex__(self):
        return self._clone_this_result(self.value.__complex__())

    def __int__(self):
        return int(self.value)
        #return self._clone_this_result(int(self.value))
        #return self._clone_this_result(self.value.__int__())

    def __float__(self):
        return self._clone_this_result(self.value.__float__())

    def __round__(self, *ndigits):
        return self._clone_this_result(self.value.__round__(*ndigits))

    def __trunc__(self):
        return math.truncate(self.value)

    def __floor__(self):
        return self._clone_this_result(self.value.__floor__())

    def __ceil__(self):
        return self._clone_this_result(self.value.__ceil__())

    def __enter__(self):
        return self.value.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        return self.value.__exit__(exc_type, exc_value, traceback)

    def __await__(self):
        return self.value.__await__()

    def __aiter__(self):
        return self.value.__aiter__()

    def __anext__(self):
        return self.value.__anext__()

    def __aenter__(self):
        return self.value.__aenter__()

    def __aexit__(self, exc_type, exc_value, traceback):
        return self.value.__aexit__(exc_type, exc_value, traceback)

    def __index__(self):
        return self.value.__index__()


def is_sandbox_result(value) -> bool:
    """
    Determines if the given value is secretly a proxied SandboxResult.
    """
    if hasattr(value, "__actual_class__"):
        if value.__actual_class__ == SandboxResult:
            return True
    return False


def len(s):
    """
    Return the length (the number of items) of an object. The argument may be
    a sequence (such as a string, bytes, tuple, list, or range) or a collection
    (such as a dictionary, set, or frozen set).

    Replacement for builtin :py:func:`len` function that is compatible
    with SandboxResult. See
    :py:func:pedal.sandbox.result.SandboxResult.__len__ for details on this
    weird thing.
    """
    if is_sandbox_result(s):
        return s._clone_this_result(_original_len(s.value))
    return len(s)


def share_sandbox_context(new_value, original_value):
    """ Produces a new version of the `new_value`, but using the sandbox,
    context ID, and metadata of the `original_value`."""
    if is_sandbox_result(original_value):
        return SandboxResult(new_value,
                             original_value._actual_context_id,
                             original_value._actual_sandbox)
    return new_value


def unwrap_value(value):
    """ Given either a SandboxResult or a regular value, make sure we have
    the actual value. """
    if is_sandbox_result(value):
        return value._actual_value
    else:
        return value




"""
The file containing the Substitution class.
"""


class Substitution:
    """
    A Substitution is a class that mocks part of the Submission class in order
    to pretend to be a different source code file. This allows Source to do a
    fancy bit of swapping around, making the code and filename appear to be
    different.
    """
    def __init__(self, code: str, filename: str):
        self.code = code
        self.filename = filename
        self.lines = code.split("\n")



"""
Classes and functions for managing the State of a variable.
"""
from pedal.types.new_types import is_subtype


def check_trace(state):
    """
    Create a list of all the types that this given state ever took on.

    Args:
        state (:py:class:`pedal.tifa.state.State`): The state to check the trace
            history of.

    Returns:
        list[:py:class:`pedal.types.new_types.Type`]: The list of all types
            that this State ever took on.
    """
    past_types = [state.type]
    for past_state in state.trace:
        past_types.extend(check_trace(past_state))
    return past_types


class State:
    """
    A representation of a variable at a particular point in time of the program.

    Attributes:
        name (str): The name of the variable, without its scope chain
        trace (list of State): A recursive definition of previous States for
                               this State.
        type (Type): The current type of this variable.
        method (str): One of 'store', 'read', (TODO). Indicates the change that
                      occurred to this variable at this State.
        position (dict): A Position dictionary indicating where this State
                         change occurred in the source code.
        read (str): One of 'yes', 'no', or 'maybe'. Indicates if this variable
                    has been read since it was last changed. If merged from a
                    diverging path, it is possible that it was "maybe" read.
        set (str): One of 'yes', 'no', or 'maybe'. Indicates if this variable
                    has been set since it was last read. If merged from a
                    diverging path, it is possible that it was "maybe" changed.
        over (str): One of 'yes', 'no', or 'maybe'. Indicates if this variable
                    has been overwritten since it was last set. If merged from a
                    diverging path, it is possible that it was "maybe" changed.
        over_position (dict): A Position indicating where the State was
                              previously set versus when it was overwritten.
    """

    def __init__(self, name, trace, type, method, position,
                 read='maybe', set='maybe', over='maybe', over_position=None):
        self.name = name
        self.trace = trace
        self.type = type
        self.method = method
        self.position = position
        self.over_position = over_position
        self.read = read
        self.set = set
        self.over = over

    def copy(self, method, position):
        """
        Make a copy of this State, copying this state into the new State's trace
        """
        return State(self.name, [self], self.type, method, position,
                     self.read, self.set, self.over, self.over_position)

    def __str__(self):
        """
        Create a string representation of this State.
        """
        return "{method}(r:{read},s:{set},o:{over},{type})".format(
            method=self.method,
            read=self.read[0],
            set=self.set[0],
            over=self.over[0],
            type=self.type.__class__.__name__
        )

    def __repr__(self):
        """
        Create a string representation of this State.
        """
        return str(self)

    def was_type(self, a_type):
        """
        Retrieve all the types that this variable took on over its entire
        trace.
        """
        from pedal.types.normalize import normalize_type
        a_type = normalize_type(a_type).as_type()
        past_types = check_trace(self)
        return any(is_subtype(past_type, a_type) for past_type in past_types)



from pedal.resolvers.core import make_resolver
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback import Feedback


@make_resolver
def resolve(report=MAIN_REPORT):
    """

    Args:
        report:

    Returns:

    """
    return ""



"""
Helper classes for storing data about executions.

NameSpace
    A dictionary that maps names to values. Represents the results of executing
    code, specifically the variables/functions/classes/etc. that exist
    afterwards. Think of it as "locals()". Represented to instructors via
    the `data` attribute.

Context
    The code that was previously executed. This is a history of what was
    executed, that can be used to augment stack traces and such. There are
    commands available to "chunk" contexts so that executions can be grouped
    as a single series of concepts.

Context ID
    A unique identifier for an execution

Execute
    Given some Python code, AST, or filename, uses ``compile`` and ``exec``
    to actually run the code.

Run
    To execute an arbitrary chunk of code, as opposed to being a ``call`` or
    ``eval``.

Call
    To execute a specific function that exists in the current NameSpace.

Eval
    To execute a chunk of code that represents an expression, storing the
    result in a temporary variable.

Target
    When code is Called or Evaled, the result is stored in a Target. This
    can be a variable, or a more complex expression.
"""
from pedal.sandbox.mocked import MockDictModule, MockModule, create_module


class SandboxContextKind:
    """ Enumeration of sandbox execution kind's. """
    RUN = 'run'
    EVAL = 'eval'
    CALL = 'call'
    GETITEM = 'getitem'

    @staticmethod
    def describe_action(context):
        """
        Produces the relevant Action message for the given kind of context.

        Args:
            context (SandboxContext): The context to describe.
        Returns:
            str: The description of the action.
        """
        if context.kind == SandboxContextKind.EVAL and context.target in ("", "_"):
            return 'I evaluated the expression'
        else:
            return 'I ran the code'

    def __repr__(self):
        return f"<SandboxContextKind {self.value}>"


class SandboxContext:
    """
    Simple data class holding information about an execution within a sandbox.
    Includes such information as the code that was run, its filename, inputs,
    outputs, exceptions, etc.

    Args:
        context_id (int): The unique ID representing this sandbox execution.
        code (str): The code that was executed.
        filename (str): The filename of the code that was executed.
        kind (SandboxContextKind): The kind of execution that generated this.
        target (str or None): If the result was stored in a variable/expression,
            this will be its name.
        inputs (list[str]): Any inputs that were used during the code's
            execution.
        output (str): The output that was generated by this execution.
        exception (Exception or None): The exception that occurred during this
            execution, or None if none occurred.
        submission (:py:class:`pedal.core.submission.Submission`): The
            submission object that this context was attached to.
        called (str): The name of the function that was called, if one was
            called.
        args (list[str]): A string representation of the arguments that were
            called.
    """

    def __init__(self, context_id, code, filename, kind, target, inputs, output,
                 exception, submission, called=None, args=None):
        self.context_id = context_id
        self.kind = kind
        # Hack: Chomp off the "_ =" if the target is "_"
        if target == "_":
            code = code[len("_ = "):]
        self.code = code
        self.filename = filename
        self.target = target
        self.inputs = inputs
        self.output = output
        self.exception = exception
        self.submission = submission
        self.called = called
        self.args = args

    def clone(self, context_id):
        """
        Create a separate copy of this item. Inputs will be deep-copied.

        Args:
            context_id (int): The new context ID.
        Returns:
            SandboxContext: The new context, copied from the old one except with a new ID.
        """
        return SandboxContext(context_id, self.code, self.filename,
                              self.kind, self.target, list(self.inputs),
                              self.output, self.exception, self.submission)

    def __repr__(self):
        return f"<SandboxContext {self.context_id}>"


class ExecutionTextHolder:
    """
    Helper class for formatting the text of an execution. Keeps track of the last action taken, the current body,
    and the current formatter for this Report.

    Args:
        format (ReportFormatter): The formatter to use for this Report.
    """
    def __init__(self, report_format):
        self._result = []
        self._last_action = None
        self._current_body = []
        self._format = report_format

    def add_message(self, action_message, body=None):
        """
        Add a message to the text of the execution.

        Args:
            action_message (str): The message to add.
            body (str): The body of the message, if any.
        """
        if action_message != self._last_action and self._last_action is not None:
            self._finish_body()
        if body:
            self._current_body.append(body)
        self._last_action = action_message

    def _finish_body(self):
        """
        Finish the current body, if any, and add it to the result.
        """
        if self._current_body:
            code_message = self._format.python_code("\n".join(self._current_body))
            message = self._last_action + "\n"+code_message + "\n"
            self._result.append(message)
            self._current_body = []
        elif self._last_action:
            self._result.append(self._last_action+"\n")

    def get_lines(self):
        """
        Get the lines of the execution text, as a single string.

        Returns:
            str: The lines of the execution text.
        """
        self._finish_body()
        return "".join(self._result)


INPUT_MAXIMUM_LINES = 30


def format_contexts(contexts, format):
    """
    Create a text string representation from a list of contexts.

    Args:
        contexts (list[list[SandboxContext]]): The list of list of sandbox
            executions. The inner list is because we could have a group of
            contexts (although more likely it'll be a single one).
        format (:py:class:`pedal.core.formatting.Formatter`): The formatter
            to use to augment the context message.

    Returns:
        str: The string representation of the contexts.
    """
    execution_text = ExecutionTextHolder(format)
    inputs_text = []
    for context_group in contexts:
        for context in context_group:
            if context.filename in (None, context.submission.instructor_file):
                action_message = SandboxContextKind.describe_action(context)
                execution_text.add_message(action_message+":", context.code)
            elif context.filename in (context.submission.main_file, ):
                execution_text.add_message(f"I ran your code.")
            else:
                filename_message = format.filename(context.filename)
                execution_text.add_message(f"I ran the file {filename_message}.")
            inputs_text.extend(context.inputs)
    final_text = [execution_text.get_lines()]
    if inputs_text:
        if len(inputs_text) > INPUT_MAXIMUM_LINES:
            midpoint = int(INPUT_MAXIMUM_LINES / 2)
            missed = len(inputs_text) - INPUT_MAXIMUM_LINES
            inputs_text = inputs_text[:midpoint] + [f"... (Skipping {missed} other inputs) ..."] + inputs_text[-midpoint:]
        inputs = "\n".join(inputs_text) + " "
        inputs_message = format.inputs(inputs)
        final_text.append(f"And I entered as input:{inputs_message}")
    return "\n".join(final_text)


class SandboxModules:
    """
    Container for any mocked modules and their data.
    """

    def __init__(self):
        self._modules = {}

    def clear(self):
        """ Removes any existing modules. """
        self._modules.clear()

    def new_module(self, data, import_name, friendly_name):
        """
        Creates a new mocked module.

        Args:
            data (dict or MockModule): The data to use for the new mocked module
            import_name (str): The name that is associated with ``import``.
            friendly_name (str): The name to use when accessing this module's
                data via attribute lookup.

        Returns:
            dict[str,:py:class:`types.ModuleType`]: The newly created modules
                mapped by their imported path names.
        """
        if isinstance(data, MockModule):
            import_name = [import_name, *data.SUBMODULES]
        root, modules, target = create_module(import_name)
        if isinstance(data, dict):
            mocked_module = MockDictModule(data)
            mocked_module.add_to_module(target)
        elif isinstance(data, MockModule):
            data.add_to_module(target, modules)
        else:
            raise ValueError("Given data must be either MockModule or dict.")
        self._modules[friendly_name] = data
        return modules

    def __getattr__(self, name):
        return self._modules[name]

    def __dir__(self):
        return self._modules.keys()


class SandboxVariable:
    """
    A representation of a variable in the student's data namespace. This
    has limited application, but can be used to improve the data passed into
    :py:func:`pedal.sandbox.commands.call` so that the local variables are used.
    Largely superceded by the ``locals_args`` parameter.
    """

    def __init__(self, name, value):
        self.name = name
        self.value = value



"""

######### Version 1

# Ends up executing all the questions' bodies
with Pool("P1") as pool:
    with Question("QA", "Create a for loop"):
        ensure_ast("For")
    with Question("QB", "Create an if statement"):
        ensure_ast("If")
pool.choose().ask()

######### Version 2

pool = Pool("P1", [question_A, question_B])

@pool.Question("Q", "Create a for loop")
def question_A():
    ensure_ast("For")

@pool.Question("QB", "Create an if statement")
def question_B():
    ensure_ast("If")

pool.choose().ask()

######### Version 3

with Pool("P1") as pool:
    if pool.asked_question("QA", "Create a for loop"):
        ensure_ast("For")
    if pool.asked_question("QB", "Create an if statement"):
        ensure_ast("If")

"""
from pedal.core.feedback import FeedbackGroup
from pedal.core.feedback_category import FeedbackCategory, FeedbackKind
from pedal.core.report import MAIN_REPORT
from pedal.questions.graders import QuestionGrader
from pedal.questions.constants import TOOL_NAME
from pedal.questions.feedbacks import show_question
from pedal.questions.pool import Pool


class QuestionGroup(FeedbackGroup):
    category = FeedbackCategory.INSTRUCTIONS
    kind = FeedbackKind.INSTRUCTIONAL
    justification = 'Question was asked but not answered.'
    title = 'Question Group'
    priority = 'instructions'
    tool = TOOL_NAME
    valence = 0
    muted = True

    def __init__(self, name, report, try_all=True, **kwargs):
        super().__init__(delay_condition=True, **kwargs)
        self.name = name
        self.all_feedback = []
        self.try_all= try_all

    def _get_child_feedback(self, feedback, active):
        self.all_feedback.append(feedback)

    def __enter__(self):
        self.report.start_group(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Go through all my collected assertions
        # Count failures, errors, successes
        # If there are any non-successes, then produce a summary table.
        #   Otherwise produce a success within this group
        # Mute and unscore any feedbacks (use their scores though)
        self.report.stop_group(self)
        self._handle_condition()
        return False


class Question:
    """

    """
    def __init__(self, name, instructions, *tests, seed=None, report=MAIN_REPORT, try_all=True):
        self.name = name
        self.instructions = instructions
        self.tests = tests
        self.seed = seed
        self.report = report
        self.answered = False
        self.try_all = try_all
        self.question_group = QuestionGroup(self.name, self.try_all)

    def __enter__(self):
        Pool.add_question_via_context(self)
        self.question_group.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.question_group.__exit__(exc_type, exc_value, traceback)

    def answer(self):
        """

        """
        self.answered = True

    def ask(self):
        """

        """
        all_tests = []
        for test in self.tests:
            # Unpack one layer of nesting
            if isinstance(test, list):
                all_tests.extend(test)
            else:
                all_tests.append(test)
        results = []
        with self.question_group:
            for test in all_tests:
                if isinstance(test, QuestionGrader):
                    results.append(test._test(self))
                else:
                    results.append(test(self))
        answered = False
        if all(r is not False for r in results) and not any(self.question_group.all_feedback):
            answered = True
        if self.answered:
            answered = True
        if not answered:
            show_question(self.instructions, report=self.report)





"""
Utilities and classes related to formatting a Feedback Message.
"""
from pedal.core.location import Location
from pedal.utilities.text import inject_line, render_table


def chomp_spec(format_spec, word):
    """ Return a new version of format_spec without the ``word`` and
    (if there are multiple format_spec given) the extra colon. """
    if format_spec.endswith(word):
        format_spec = format_spec[:-len(word)]
        if format_spec and format_spec[-1] == ':':
            format_spec = format_spec[:-1]
    return format_spec


def escape_dunders(text):
    if text[:2] == '__' and text[-2:] == '__':
        return "\\_\\_" + text[2:-2] + "\\_\\_"
    return text


def escape_angles(text):
    return text.replace("<", "&lt;").replace(">", "&gt;")


class FeedbackFieldWrapper:
    """
    Wraps an individual field within a Feedback message.

    Args:
        key (str): The name of the field.
        value (Any): The value to interpolate into the message.
        formatter (Formatter): The formatter to use from the report.
    """
    def __init__(self, key, value, formatter):
        self.key = key
        self.value = value
        self.formatter = formatter

    def __getattr__(self, key):
        return FeedbackFieldWrapper(self.key, getattr(self.value, key), self.formatter)

    def __getitem__(self, index):
        return FeedbackFieldWrapper(self.key, self.value[index], self.formatter)

    def __repr__(self):
        return repr(self.value)

    def __str__(self):
        return str(self.value)

    def __format__(self, format_spec):
        value = str(self.value)
        for formatter_name in self.formatter.available:
            if format_spec.endswith(formatter_name):
                format_spec = chomp_spec(format_spec, formatter_name)
                value = getattr(self.formatter, formatter_name)(self.value)
                break
        return value.__format__(format_spec)

# TODO: Convert Formatter into HtmlFormatter, and then make Formatter text by
#       default.


def wrap_fields(formatter, fields):
    return {field: FeedbackFieldWrapper(field, value, formatter) for field, value in fields.items()}


class Formatter:
    """ Utility class for wrapping a feedback message with HTML or other
            extra information. Can be subclassed by a different environment and
            attached to a Report, in order to change how we create feedback. """

    available = ['exception', 'filename', 'frame', 'traceback',
                 # 'html_code', 'html_div', 'html_pre', 'html_span', 'html_tag',
                 'inputs', 'line', 'name', 'output',
                 'python_code', 'python_expression', 'python_value',
                 'table']

    def __init__(self, report=None):
        """ Can optionally specify a report, in order to look over current
        feedback and settings. """
        self.report = report

    def update_report(self, report):
        """ Change the currently set report for this formatter. """
        self.report = report

    def pre(self, text):
        return "\n".join(["    " + line for line in text.split("\n")])

    def python_code(self, code, focus=None):
        if isinstance(focus, Location):
            code = inject_line(code, focus.line+1, " "*(focus.col-1) + "^"*(focus.end_col-focus.col))
        return self.pre(code)

    def python_expression(self, code):
        return str(code)

    def filename(self, filename):
        return filename

    def python_value(self, code):
        return self.pre(code)

    def inputs(self, inputs):
        return "\n"+self.pre(inputs)+"\n"

    def output(self, output):
        return self.pre(output)

    def traceback(self, traceback):
        return traceback

    def name(self, name):
        return name

    def line(self, line_number):
        return line_number

    def frame(self, name):
        return name

    def exception(self, exception):
        return self.pre(exception)

    def table(self, rows, columns):
        return "\n" + render_table(rows, columns, " | ")

    def check_mark(self):
        # TODO: should be ✓, right?
        return " "

    def negative_mark(self):
        return "×"

    def indent(self):
        return "    "


class HtmlFormatter(Formatter):

    def html_tag(self, tag, contents, classes=None):
        if classes is None:
            return f"<{tag}>{contents}</{tag}>"
        else:
            return f"<{tag} class='{classes}'>{contents}</{tag}>"

    def html_pre(self, text, classes=None):
        return self.html_tag('pre', escape_angles(text), classes)

    def html_div(self, contents, classes=None):
        return self.html_tag('div', contents, classes)

    def html_code(self, contents, classes=None):
        return self.html_tag('code', escape_dunders(contents), classes)

    def html_span(self, contents, classes=None):
        return self.html_tag('span', contents, classes)

    ############################################################################

    def python_code(self, code, focus=None):
        if isinstance(focus, Location):
            code = inject_line(code, focus.line+1, " "*(focus.col-1) + "^"*(focus.end_col-focus.col))
        return self.html_tag('pre', self.html_code(code), "pedal-python-code python")

    def python_expression(self, code):
        return self.html_code(code)

    def filename(self, filename):
        return self.html_code(filename, "pedal-filename")

    def python_value(self, code):
        return self.html_pre(code, "pedal-python-value")

    def inputs(self, inputs):
        return self.html_pre(inputs, "pedal-inputs")

    def output(self, output):
        return self.html_pre(output, "pedal-output")

    def traceback(self, traceback):
        return self.html_div(traceback, "pedal-traceback")

    def name(self, name):
        return self.html_code(name, "pedal-name")

    def line(self, line_number):
        return str(line_number)

    def frame(self, name):
        return self.html_code(name, "pedal-frame")

    def exception(self, exception):
        return self.html_pre(exception, "pedal-exception")

    def table(self, rows, columns):
        """
        Creates a tabular representation of the rows and columns.
        If a row's length is less than the columns' length, then it will
        receive a colspan to stretch its last cell to fit the table.

        Args:
            rows (list[list[str]]): The 2D cells of the table.
            columns (list[str]): The header rows

        Returns:
            str: The completed table.
        """
        cell_class = " class='pedal-cell'"
        row_class = " class='pedal-row'"
        table_class = " class='pedal-table'"
        header_class = " class='pedal-header'"
        header = "\n".join(f"    <th{cell_class}>{column}</th>" for column in columns)
        body = []
        for row in rows:
            # Skip empty rows
            if not row:
                continue
            row_text = []
            # If we miss some columns, expand the last one.
            length_difference = max(len(columns) - len(row) + 1, 0)
            colspan = " colspan={length_difference}" if length_difference > 1 else ""
            # Make the cells, but we'll treat the last one as special
            for cell in row[:-1]:
                row_text.append(f"    <td{cell_class}>{cell}</td>")
            #  Add in the colspan to the last one
            row_text.append(f"    <td{colspan}{cell_class}>{row[-1]}</td>")
            row_text = "\n".join(row_text)
            body.append(f"  <tr{row_class}>\n{row_text}\n  </tr>")
        body = "\n".join(body)
        return (f"\n<table{table_class}>"
                f"   <tr{header_class}>\n{header}\n  </tr>"
                f"   {body}"
                f"</table>")

    def indent(self):
        return "&nbsp;"*4

    def check_mark(self):
        return self.html_span("&#10004;", "pedal-positive-mark")

    def negative_mark(self):
        return self.html_span("&#10060;", "pedal-negative-mark")


class TextFormatter(Formatter):
    pass



"""
Tests related to checking function definitions
"""

import unittest

from pedal.utilities.system import IS_AT_LEAST_PYTHON_311
from pedal import unit_test, block_function, evaluate, CommandBlock, run, call
from pedal.assertions.static import *
from pedal.assertions.runtime import assert_equal, assert_greater, assert_output_contains
from pedal.types.new_types import DictType, LiteralStr, StrType
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE


class TestAssertions(ExecutionTestCase):
    def test_function_found_complex_passes(self):
        with Execution('def x(a: int, b: str, c: [int]) -> str: pass\nx') as e:
            self.assertFalse(ensure_function('x', 3, (int, str, [int]),
                                             str))
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_function_found_complex_fails(self):
        with Execution('def x(a: int, b: str, c: [int]) -> str: pass\nx') as e:
            self.assertTrue(ensure_function('x', 3, (int, str, [str]),
                                             str))
        self.assertFeedback(e, """Wrong Parameter Type
The function named x has a parameter named c that is a list of integers, but should be a list of strings.""")

    def test_function_missing_parameter(self):
        with Execution('def x(a: int, b, c: [int]) -> str: pass\nx') as e:
            self.assertTrue(ensure_function('x', 3, (int, str, 'list[int]'), str))
        self.assertFeedback(e, "Missing Parameter Type\n"
                               "The function named x "
                               "has a parameter named b, "
                               "but that parameter does not have a type specified.")

    @unittest.skip("This is currently broken!")
    def test_function_custom_class(self):
        with Execution('D = list\ndef x(a: int, c: D) -> str: pass\nx\nD') as e:
            f = ensure_function('x', parameters=(int, 'int'), returns=str)
            self.assertTrue(f)
        #print(f.fields)
        self.assertFeedback(e, "Wrong Parameter Type\n"
                               "The function named x "
                               "has a parameter named c "
                               "that is a 'D' type, but should be a number.")

    def test_assert_type_custom_record_passes(self):
        Dog = DictType({LiteralStr("Name"): StrType()})
        with Execution('Dog = {"Name": str}\ndef pet(d: Dog) -> str: return d["Name"]\npet({"Name": "Fido"})') as e:
            f = ensure_function('pet', parameters=('Dog',), returns=str)
            self.assertFalse(f)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_assert_type_custom_record_fails(self):
        Dog = DictType({LiteralStr("Name"): StrType()})
        with Execution('def pet(d: str) -> str: return d\npet("Fido")') as e:
            f = ensure_function('pet', parameters=('Dog',), returns=str)
            self.assertTrue(f)
        self.assertFeedback(e, "Wrong Parameter Type\n"
                               "The function named pet has a parameter named d"
                               " that is a string, but should be A Dog.")

    def test_unit_test_partial_credit_true(self):
        with Execution('def add(a, b): return b', run_tifa=False) as e:
            unit_test('add',
                      ((1, 2), 3),
                      ((0, 0), 0),
                      ((0, 3), 3),
                      ((0, 5), 5),
                      ((1, 3), 4),
                      score="+100%",
                      partial_credit=True)
        self.assertEqual(3/5, e.final.score)
        self.assertFeedback(e, """
Failed Instructor Test
Student code failed instructor tests.
You passed 3/5 tests.

I ran your function add on some new arguments.
  | Arguments | Returned | Expected
===================================
× |      1, 2 |        2 |        3
  |      0, 0 |        0 |        0
  |      0, 3 |        3 |        3
  |      0, 5 |        5 |        5
× |      1, 3 |        3 |        4""")

    def test_bad_name_error_concatenation(self):
        with Execution("""
def make_polite(message: str) -> str:
    return ", please"+make_polite

make_polite("Pet the dog")""", run_tifa=False) as e:
            unit_test('make_polite',
                      ([""], ", please"),
                      (["Pet the dog"], "Pet the dog, please"),
                      (["Walk the dog"], "Walk the dog, please"))
        self.assertFeedback(e, """Type Error
I ran your code.

A TypeError occurred:

    Can only concatenate str (not "function") to str

The traceback was:
Line 5 of file answer.py
    make_polite("Pet the dog")
""" + ('    ^^^^^^^^^^^^^^^^^^^^^^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + """
Line 3 of file answer.py in make_polite
        return ", please"+make_polite
""" + ('               ^^^^^^^^^^^^^^^^^^^^^^\n' if IS_AT_LEAST_PYTHON_311 else "") + """
Type errors occur when you use an operator or function on the wrong type of value. For example, using `+` to add to a list (instead of `.append`), or dividing a string by a number.

Suggestion: To fix a type error, you should trace through your code. Make sure each expression has the type you expect it to have.""")

    def test_unit_test_with_style(self):
        with Execution("""
def add(a, b):
    return a - b""", run_tifa=False) as e:
            with unit_test('add'):
                assert_equal(call('add', 1, 2), 3)
                assert_equal(call('add', 0, 0), 0)
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor tests.
You passed 1/2 tests.

I ran your function add on some new arguments.
  | Arguments | Returned | Expected
===================================
× |      1, 2 |       -1 |        3
  |      0, 0 |        0 |        0""")

    def test_unit_test_blocked_function(self):
        with Execution("""
def summate(numbers: [int]) -> int:
    return sum(numbers)

print(summate([1,2,3]))""", run_tifa=False) as e:
            block_function('sum')
            unit_test('summate', [([1, 2, 3],), 6])
        self.assertFeedback(e, """Failed Instructor Test
Student code failed instructor tests.
You passed 0/1 tests.

I ran your function summate on some new arguments.
  |     Arguments |                           Returned | Expected
=================================================================
× |     [1, 2, 3] | You are not allowed to call 'sum'. |        6""")

    def test_unit_test_extra_context(self):
        with Execution('def add(a, b): return b', run_tifa=False) as e:
            unit_test('add',
                      ('data, 5', 17),
                      (([3, 3], 3), 9),
                      score="+100%",
                      context=evaluate('[5, 3, 4]', target='data'),
                      partial_credit=True)
        #self.assertEqual(3 / 5, e.final.score)
        self.assertFeedback(e, """
Failed Instructor Test
Student code failed instructor tests.
You passed 0/2 tests.

I ran the code:
    data = [5, 3, 4]

I ran your function add on some new arguments.
  |     Arguments | Returned | Expected
=======================================
× |       data, 5 |        5 |       17
× |     [3, 3], 3 |        3 |        9""")

    def test_unit_test_extra_context_run(self):
        with Execution('def add(a, b): return b', run_tifa=False) as e:
            run('this_should_not_appear = 5')
            unit_test('add',
                      ('data, 5', 17),
                      (([3, 3], 3), 9),
                      score="+100%",
                      context=run('data = [5, 3, 4]\ndata2 = [1,2,3]'),
                      partial_credit=True)
        #self.assertEqual(3 / 5, e.final.score)
        self.assertFeedback(e, """
Failed Instructor Test
Student code failed instructor tests.
You passed 0/2 tests.

I ran the code:
    data = [5, 3, 4]
    data2 = [1,2,3]

I ran your function add on some new arguments.
  |     Arguments | Returned | Expected
=======================================
× |       data, 5 |        5 |       17
× |     [3, 3], 3 |        3 |        9""")

    def test_unit_test_extra_context_calls(self):
        with Execution('def add(a, b): return b', run_tifa=False) as e:
            with CommandBlock() as context:
                evaluate('[1,2,3]', target='data')
                evaluate('[1,2,3]', target='data2')
            unit_test('add',
                      ('data, 5', 17),
                      ('data2, 5', 17),
                      (([3, 3], 3), 9),
                      score="+100%",
                      context=context,
                      partial_credit=True)
        #self.assertEqual(3 / 5, e.final.score)
        self.assertFeedback(e, """
Failed Instructor Test
Student code failed instructor tests.
You passed 0/3 tests.

I ran the code:
    data = [1,2,3]
    data2 = [1,2,3]

I ran your function add on some new arguments.
  |     Arguments | Returned | Expected
=======================================
× |       data, 5 |        5 |       17
× |      data2, 5 |        5 |       17
× |     [3, 3], 3 |        3 |        9""")

if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Context managers for dealing with paths and scopes.
"""


class NewPath:
    """
    Context manager for entering and leaving execution paths (e.g., if
    statements).

    Args:
        tifa (Tifa): The tifa instance, so we can modify some of its
                     properties that track variables and paths.
        origin_path (int): The path ID parent to this one.
        name (str): The symbolic name of this path, typically 'i' for an IF
                    body and 'e' for ELSE body.

    Fields:
        id (int): The path ID of this path
    """

    def __init__(self, tifa, origin_path, name):
        self.tifa = tifa
        self.name = name
        self.origin_path = origin_path
        self.id = None

    def __enter__(self):
        self.tifa.path_id += 1
        self.id = self.tifa.path_id
        self.tifa.path_names.append(str(self.id) + self.name)
        self.tifa.path_chain.insert(0, self.id)
        self.tifa.name_map[self.id] = {}
        self.tifa.path_parents[self.id] = self.origin_path

    def __exit__(self, exc_type, value, traceback):
        self.tifa.path_names.pop()
        self.tifa.path_chain.pop(0)


class NewScope:
    """
    Context manager for entering and leaving scopes (e.g., inside of
    function calls, class bodies).

    Args:
        tifa (Tifa): The tifa instance, so we can modify some of its
            properties that track variables and paths.
        definitions_scope_chain (list of int): The scope chain of the
            definition.
        class_type (ClassType): If this scope is a ClassType, then this
            will be available as a field.
    """

    def __init__(self, tifa, definitions_scope_chain, class_type=None, is_module=False):
        self.tifa = tifa
        self.definitions_scope_chain = definitions_scope_chain
        self.class_type = class_type
        self.is_module = is_module

    def __enter__(self):
        # Manage scope
        self.old_scope = self.tifa.scope_chain[:]
        # Move to the definition's scope chain
        self.tifa.scope_chain = self.definitions_scope_chain[:]
        # And then enter its body's new scope
        self.tifa.scope_id += 1
        self.tifa.scope_chain.insert(0, self.tifa.scope_id)
        # Register as class potentially
        if self.class_type is not None:
            self.class_type.scope_id = self.tifa.scope_id
            self.tifa.class_scopes[self.tifa.scope_id] = self.class_type
        if self.is_module:
            self.tifa.module_scopes[self.tifa.scope_id] = self.class_type

    def __exit__(self, exc_type, value, traceback):
        # Finish up the scope
        if not self.is_module:
            self.tifa._finish_scope()
        # Leave the body
        self.tifa.scope_chain.pop(0)
        # Restore the scope
        self.tifa.scope_chain = self.old_scope



"""
Tools are effectively submodules within Pedal - notable exceptions are
:py:mod:`pedal.core`, :py:mod:`pedal.environments`,
:py:mod:`pedal.utilities`, and :py:mod:`pedal.resolvers`.

All Tools with any kind of state are expected to have a `reset` function.
Although this can take parameters, we recommend avoiding that. The `reset`
should be for putting things back into a "null" state, and then you could have
followup functions that also give initial state.

Tools should avoid removing their references (i.e., clear out data from lists
and dictionaries instead of assigning a new one).

Tools should define a ``constants.py`` file with any useful constants. One of
these should be ``TOOL_NAME``, a string value indicating their desired namespace
(e.g., ``TOOL_NAME = "source"``).

Tools should define a ``feedbacks.py`` module that centralizes
:term:`Feedback Functions<Feedback Function>` for that tool.

Any function that interacts with Reports should expose a ``report`` parameter
that defaults to :py:data:`pedal.core.report.MAIN_REPORT`. All internal
functions should respect the ``report`` that was passed in, and not assume the
``MAIN_REPORT``.

Tools are allowed to store state within their namespace of a
:py:class:`pedal.core.report.Report`.
If the tool has not yet been initialized, its `reset` function will be called.
You can update and access fields via dictionary access.

.. code-block:: python

    report[TOOL_NAME]['my field'] = False

The ``__init__.py`` file for a Tool should use  ``__all__`` to expose any
interesting teacher-level functions, classes, and data. That way, teachers can
just consistently use ``from pedal.tool import *`` to instantly gain access to
the interesting set of members.

"""

__all__ = ['ToolRegistration']


class ToolRegistration:
    """
    ToolRegistration is a data class for holding general, system-wide
    information about a Tool. Note that it doesn't hold the tool's data, just
    the static information about the tool like its reset function.

    Attributes:
        name (str): The formal name of this tool.
        reset (callable): A function defined for initializing and
            re-initializing the tool's data.
    """
    def __init__(self, name, reset):
        self.name = name
        self.reset = reset



"""
Check pedal.types
"""
import ast
import unittest
import pedal.types.new_types as types
from pedal.types.new_types import is_subtype
from pedal.types.normalize import normalize_type, get_pedal_type_from_value


class TestNormalizeType(unittest.TestCase):
    def assert_is_subtype(self, left, right):
        self.assertTrue(is_subtype(left, right.as_type()),
                        f"{left.singular_name} is (incorrectly) not a subtype of {right.singular_name}")

    def assert_is_not_subtype(self, left, right):
        self.assertFalse(is_subtype(left, right.as_type()),
                        f"{left.singular_name} is (incorrectly) a subtype of {right.singular_name}")

    def test_subtyping(self):
        self.assert_is_subtype(types.LiteralInt(5), normalize_type(int))
        self.assert_is_subtype(get_pedal_type_from_value(5), normalize_type(int))
        self.assert_is_subtype(get_pedal_type_from_value([1, 2]), normalize_type(list))
        self.assert_is_not_subtype(get_pedal_type_from_value({1:2}), normalize_type(dict))

    def test_simple_primitives(self):
        self.assertIsInstance(normalize_type(int), types.IntType)
        self.assertIsInstance(normalize_type(bool), types.BoolType)
        self.assertIsInstance(normalize_type(float), types.FloatType)
        self.assertIsInstance(normalize_type(str), types.StrType)
        self.assertIsInstance(normalize_type(list), types.ListConstructor)
        self.assertIsInstance(normalize_type(type(None)), types.NoneType)

    def test_complex_annotation(self):
        list_int = normalize_type('list[int]')
        self.assertIsInstance(list_int, types.ListConstructor)
        self.assertIsInstance(list_int.type_arguments, types.IntType)

        bad_dict = normalize_type(ast.parse('dict[int]'))
        self.assertIsInstance(bad_dict, types.ImpossibleType)

    def test_literals_normalized(self):
        self.assertIsInstance(normalize_type(5), types.LiteralInt)


if __name__ == '__main__':
    unittest.main(buffer=False)



"""
Exceptions related to Sandbox execution; note that these are not meant for
students, they indicate system errors.
"""


class SandboxException(Exception):
    """
    Generic base exception for sandbox errors.
    """


class SandboxStudentCodeException(SandboxException):
    """
    Caused by an error in student code
    """
    def __init__(self, actual):
        self.actual = actual


class SandboxPreventModule(Exception):
    """
    Caused by student attempting to load a module that they shouldn't.
    """


class SandboxHasNoFunction(SandboxException):
    """
    Caused by attempting to access a function that the student hasn't created.
    """


class SandboxHasNoVariable(SandboxException):
    """
    Caused by attempting to access a variable that the student hasn't created.
    """


class SandboxNoMoreInputsException(Exception):
    """
    Caused by the student calling `input` when the instructor hasn't provided
    enough inputs. Typically, the student has an infinite loop around their
    `input` function.
    """



"""
A version of the resolver that still chooses one piece of feedback, but
does so per section.
"""

from pedal.core.feedback import Feedback
from pedal.core.final_feedback import FinalFeedback, set_correct_no_errors
from pedal.core.report import MAIN_REPORT
from pedal.resolvers.core import make_resolver
from pedal.resolvers.simple import by_priority


@make_resolver
def resolve(report=MAIN_REPORT, priority_key=by_priority):
    """
    Args:
        priority_key: The key function to sort feedbacks by
        report (Report): The report object to resolve down. Defaults to the
                         global MAIN_REPORT

    Returns
        str: A string of HTML feedback to be delivered
    """
    # Prepare feedbacks
    all_feedbacks = report.feedback
    # Group the feedback
    feedback_by_group = {}
    for feedback in all_feedbacks:
        if feedback.parent in feedback_by_group:
            feedback_by_group[feedback.parent].append(feedback)
        else:
            feedback_by_group[feedback.parent] = [feedback]
    # Process each subgroup
    finals = {}
    for group, feedbacks in feedback_by_group.items():
        feedbacks.sort(key=priority_key)
        # Create the initial final feedback
        final = set_correct_no_errors(report)
        # Process each feedback in turn
        for feedback in feedbacks:
            final.merge(feedback)
        # Override empty message
        final.finalize()
        finals[group] = final
    report.result = finals
    report.resolves.append(finals)
    return finals



"""
instructions: blah blah blah

settings:
    tifa:
        enabled: True
    unit test by function (bool): Whether to test each function entirely before moving onto the
            next one, or to first check that all functions have been defined, and then
            checking their parameters, etc. Defaults to True.
    show case details (bool): Whether to show the specific args/inputs that caused a test case
            to fail.
rubric:
    functions:
        total: 100
        definition: 10
        signature: 10
        cases: 80
global:
    variables:
        name:
        type:
        value:
    inputs:
    prints:
# Sandbox, type checking
functions:
    documentation: "any" or "google"
    coverage: 100%
    tests: int
    name: do_complicated_stuff
    arity: int
    signature: int, int -> float
    signature: int, int, list[int], (int->str), dict[str:list[int]] -> list[int]
    parameters:
        name: banana
            exactly:
            regex:
            includes:
            within:
        type: int
    cases:
      - arguments (list): 5, 4
        inputs (list):
        returns (Any):
            equals: 27.3
            is:
            is not: _1
        name (str): Meaningful name for tracking purposes? Or possibly separate into label/id/code
        hint (str): Message to display to user
        prints:
            exactly:
            regex:
            startswith:
            endswith:
        plots:
# Cait
syntax:
    prevent:
        ___ + ___
# Override any of our default feedback messages
messages:
    FUNCTION_NOT_DEFINED: "Oops you missed a function"
"""
from pedal.core.commands import set_correct, give_partial, gently
from pedal.core.feedback_category import FeedbackCategory
from pedal.questions.constants import TOOL_NAME

from pedal.sandbox.commands import get_sandbox, get_student_data
from pedal.utilities.comparisons import equality_test
from pedal.assertions.static import ensure_function
from pedal.assertions.runtime import ensure_function_callable
from pedal.assertions.commands import unit_test

SETTING_SHOW_CASE_DETAILS = "show case details"
DEFAULT_SETTINGS = {
    SETTING_SHOW_CASE_DETAILS: True
}

EXAMPLE_DATA = {
    'functions': [{
        'name': 'do_complicated_stuff',
        'signature': 'int, int, [int] -> list[int]',
        'cases': [
            {'arguments': "5, 4, 3", 'returns': "12"},
        ]
    }]
}


def load_question(data):
    """

    :param data:
    :return:
    """
    settings = DEFAULT_SETTINGS.copy()
    settings.update(data.get('settings', {}))
    rubric = settings.get('rubric', {})
    if 'functions' in data:
        function_rubric = rubric.get('functions', {})
        for function in data['functions']:
            ensure_function(function['name'], arity=function.get('arity'),
                            parameters=function.get('parameters'),
                            returns=function.get('returns'),
                            score=function_rubric.get('definition', 10))
            ensure_function_callable(function['name'])
            unit_test(function, *function.get('cases', []))

def check_question(data):
    """

    Args:
        data:
    """
    results = list(load_question(data))
    if results:
        message, label = results[0]
        gently(message, label=label)


def check_pool(questions):
    """

    Args:
        questions:
    """
    pass


def load_file(filename):
    """

    Args:
        filename:
    """
    pass
